# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pts.dataset.repository.datasets import get_dataset, dataset_recipes
from pts.dataset.utils import to_pandas
import numpy as np
import pandas as pd
from pts.dataset import ListDataset
import torch
from pts.model.simple_feedforward import SimpleFeedForwardEstimator
import pickle
import json
import random


def get_m4_by_freq(
    context_length=72,
    prediction_length=24,
    len_per_ts=200,
    num_ts=50,
    num_groups=6,
    file_name="m4_freq",
):
    dataset_group = [[] for i in range(num_groups)]
    whole_data = []
    ret = dict()
    datasets_name = [
        "m4_hourly",
        "m4_daily",
        "m4_weekly",
        "m4_monthly",
        "m4_quarterly",
        "m4_yearly",
    ]
    hours_factor = [
        1,
        24,
        24 * 7,
        24 * 7 * 30,
        24 * 7 * 30 * 3,
        24 * 7 * 30 * 3 * 4,
    ]
    for i in range(num_groups):
        dataset = get_dataset(datasets_name[i])
        len_sample = context_length + prediction_length
        it = iter(dataset.train)
        for j in range(num_ts):
            train_entry = next(it)
            unsplit_ts = train_entry["target"]
            # unsplit_start = train_entry['start']
            unsplit_start = pd.Timestamp("1990-01-01")
            for ts_sample_start in range(
                0, len_per_ts - len_sample, prediction_length
            ):
                if len_sample > len(unsplit_ts):
                    continue
                ts_slice = unsplit_ts[
                    ts_sample_start : ts_sample_start + len_sample
                ]
                if len(ts_slice) < len_sample:
                    continue
                nu = 1 + sum(ts_slice) / len_sample
                ts_slice = [i / nu for i in ts_slice]
                whole_data.append(
                    {
                        "target": ts_slice,
                        "start": unsplit_start,
                        "feat_static_cat": train_entry["feat_static_cat"],
                    }
                )
                dataset_group[i].append(
                    {
                        "target": ts_slice,
                        "start": unsplit_start,
                        "feat_static_cat": train_entry["feat_static_cat"],
                    }
                )
                # unsplit_start += pd.Timedelta(hours=prediction_length*hours_factor[i])
                unsplit_start += pd.Timedelta(hours=prediction_length)
    # for j in range(len(dataset_group)):
    #    print(len(dataset_group[i]))
    # import pdb;pdb.set_trace()
    print(len(whole_data))
    ret["group_ratio"] = [len(i) / len(whole_data) for i in dataset_group]
    print(ret["group_ratio"])
    random.shuffle(whole_data)
    ret["whole_data"] = ListDataset(whole_data, freq=dataset.metadata.freq)
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq=dataset.metadata.freq))
    ret["group_data"] = group_data_list
    print("write whole data")
    with open("synthetic_" + file_name + "_whole_data.csv", "wb") as output:
        pickle.dump(ret["whole_data"], output)
    print("write group data")
    with open("synthetic_" + file_name + "_group_data.csv", "wb") as output:
        pickle.dump(ret, output)
    return True


def get_temperature_data(
    context_length=24, prediction_length=4, samples_per_ts=2000, num_groups=8
):
    ts_file = pd.read_csv("temperature.csv")
    city_names = [
        "Vancouver",
        "Los Angeles",
        "Las Vegas",
        "San Diego",
        "Philadelphia",
        "Montreal",
        "Boston",
        "Haifa",
    ]
    datetime = ts_file["datetime"]
    dataset_group = [[] for i in range(num_groups)]
    whole_data = []
    ret = dict()
    for gid in range(num_groups):
        ts = ts_file[city_names[gid]]
        num_samples = 0
        index = 0
        while True:
            num_samples += 1
            index += 1
            ts_slice = torch.tensor(
                ts[index : index + context_length + prediction_length].values
            )
            nu = 1 + sum(ts_slice) / len(ts_slice)
            ts_slice /= nu
            if torch.sum(torch.isnan(ts_slice)).item() == 0:
                dataset_group[gid].append(
                    {
                        "target": ts_slice,
                        "start": pd.Timestamp(datetime[index]),
                    }
                )
                whole_data.append(
                    {
                        "target": ts_slice,
                        "start": pd.Timestamp(datetime[index]),
                    }
                )
            if num_samples == samples_per_ts:
                break
    random.shuffle(whole_data)
    ret["whole_data"] = ListDataset(whole_data, freq="1H")
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq="1H"))
    ret["group_data"] = group_data_list
    print("write whole data")
    with open("synthetic_temperature_whole_data.csv", "wb") as output:
        pickle.dump(ret["whole_data"], output)
    print("write group data")
    with open("synthetic_temperature_group_data.csv", "wb") as output:
        pickle.dump(ret, output)
    return True


def get_amazon_sales():
    f = open("./dataset/sgc_train.json", encoding="utf-8")
    dataset_group = [[] for i in range(8)]
    whole_data = []
    ret = dict()
    X = []
    Y = []
    # split_grid = [0.04, 0.1, 0.25, 0.5, 1, 10, 100, 5000]
    for line in f.readlines():
        # gid = 0
        dic = json.loads(line)
        var = torch.var(torch.FloatTensor(dic["target"])).item()
        if var > 5000:
            continue
        nu = 1 + sum(dic["target"]) / len(dic["target"])
        ts = [i / nu for i in dic["target"]]
        start = dic["start"]
        # ts = dic['target']
        if len(ts) < 28:
            continue
        X.append((ts, var))
    X = sorted(X, key=lambda x: x[1])
    X = [x[0] for x in X]
    length = int(len(X) / 8)
    for gid in range(8):
        for j in range(gid * length, (gid + 1) * length):
            whole_data.append({"target": X[j], "start": start})
            dataset_group[gid].append({"target": X[j], "start": start})
    random.shuffle(whole_data)
    ret["whole_data"] = ListDataset(whole_data, freq="1H")
    group_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_list.append(ListDataset(group, freq="1H"))
    ret["group_data"] = group_list
    print("write whole data")
    with open("synthetic_sales_time_whole_data.csv", "wb") as output:
        pickle.dump(ret["whole_data"], output)
    print("write group data")
    with open("synthetic_sales_time_group_data.csv", "wb") as output:
        pickle.dump(ret, output)
    return True


def get_group_data_by_var(name, num_groups, len_sample=9):
    dataset = get_dataset(name)
    dataset_group = [[] for i in range(num_groups)]
    whole_data = []
    ret = []
    it = iter(dataset.train)
    num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    group_boundary = [1e3, 5e3, 1e4, 5e4, 1e5, 5e5]
    for i in range(num_ts):
        train_entry = next(it)
        unsplit_ts = train_entry["target"][0:800]
        unsplit_start = train_entry["start"]
        whole_data.append({"target": unsplit_ts, "start": unsplit_start})
        for ts_sample_start in range(len(unsplit_ts) - len_sample):
            group_id = 0
            print(
                torch.var(
                    torch.FloatTensor(
                        unsplit_ts[
                            ts_sample_start : ts_sample_start + len_sample
                        ]
                    )
                )
            )
            continue
            dataset_group[group_id].append(
                {
                    "target": unsplit_ts[
                        ts_sample_start : ts_sample_start + len_sample
                    ],
                    "start": unsplit_start,
                }
            )
            unsplit_start += pd.Timedelta(hours=1)
    import pdb

    pdb.set_trace()
    random.shuffle(whole_data)
    print("append once")
    ret.append(ListDataset(whole_data, freq=dataset.metadata.freq))
    print("append twice")
    ret.append(ListDataset(whole_data, freq=dataset.metadata.freq))
    print("append data")
    for group in dataset_group:
        random.shuffle(group)
        ret.append(ListDataset(group, freq=dataset.metadata.freq))
    print("write whole data")
    with open("synthetic_traffic_time_whole_data.csv", "wb") as output:
        pickle.dump(ret[0:2], output)
    print("write group data")
    with open("synthetic_traffic_time_group_data.csv", "wb") as output:
        pickle.dump(ret, output)
    return True


def get_group_data_by_hash(name, q, num_groups):
    dataset = get_dataset(name)
    dataset_group = [[] for i in range(num_groups)]
    whole_data = []
    ret = []
    it = iter(dataset.train)
    num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    for i in range(num_ts):
        train_entry = next(it)
        dataset_group[i % num_groups].append(
            {"target": train_entry["target"], "start": train_entry["start"]}
        )
        whole_data.append(
            {"target": train_entry["target"], "start": train_entry["start"]}
        )
    random.shuffle(whole_data)
    ret.append(ListDataset(whole_data, freq=dataset.metadata.freq))
    for group in dataset_group:
        ret.append(ListDataset(group, freq=dataset.metadata.freq))
    return ret, dataset.metadata.freq


def get_group_data_by_duplicate(name, num_duplicates, num_groups):
    dataset = get_dataset(name)
    dataset_group = [[] for i in range(num_groups)]
    whole_data_list = []
    no_duplicate_whole_data_list = []
    ret = []
    it = iter(dataset.train)
    num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    for i in range(num_ts):
        train_entry = next(it)
        no_duplicate_whole_data_list.append(
            {"target": train_entry["target"], "start": train_entry["start"]}
        )
        for j in range(num_duplicates):
            dataset_group[i % num_groups].append(
                {
                    "target": train_entry["target"],
                    "start": train_entry["start"],
                }
            )
            whole_data_list.append(
                {
                    "target": train_entry["target"],
                    "start": train_entry["start"],
                }
            )
    random.shuffle(whole_data_list)
    random.shuffle(no_duplicate_whole_data_list)
    ret.append(
        ListDataset(no_duplicate_whole_data_list, freq=dataset.metadata.freq)
    )
    ret.append(ListDataset(whole_data_list, freq=dataset.metadata.freq))
    for group in dataset_group:
        random.shuffle(group)
        ret.append(ListDataset(group, freq=dataset.metadata.freq))
    return ret, dataset.metadata.freq


def get_whole_data_by_duplicate(name, num_duplicates):
    dataset = get_dataset(name)
    dataset_group = []
    ret = []
    no_duplicate_whole_data_list = []
    it = iter(dataset.train)
    num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    for i in range(num_ts):
        train_entry = next(it)
        no_duplicate_whole_data_list.append(
            {"target": train_entry["target"], "start": train_entry["start"]}
        )
        for j in range(num_duplicates):
            dataset_group.append(
                {
                    "target": train_entry["target"],
                    "start": train_entry["start"],
                }
            )
    random.shuffle(dataset_group)
    random.shuffle(no_duplicate_whole_data_list)
    ret.append(
        ListDataset(no_duplicate_whole_data_list, freq=dataset.metadata.freq)
    )
    ret.append(ListDataset(dataset_group, freq=dataset.metadata.freq))
    return ret, dataset.metadata.freq


def get_group_data(name):
    dataset = get_dataset(name)
    dataset_group = []
    it = iter(dataset.train)
    num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    for i in range(num_ts):
        train_entry = next(it)
        dataset_group.append(
            ListDataset(
                [
                    {
                        "target": train_entry["target"],
                        "start": train_entry["start"],
                    }
                ],
                freq=dataset.metadata.freq,
            )
        )
    return dataset_group


def get_whole_data(name):
    dataset = get_dataset(name)
    dataset_group = []
    it = iter(dataset.train)
    num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    for i in range(num_ts):
        train_entry = next(it)
        dataset_group.append(
            {"target": train_entry["target"], "start": train_entry["start"]}
        )
    return ListDataset(dataset_group, freq=dataset.metadata.freq)


def get_synthetic_data(model_name=None, num_groups=8, mean_boundary=1):
    assert num_groups > 1
    prediction_length = 1
    context_length = 5
    num_time_steps = 1
    # device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    device = "cpu"
    net = SimpleFeedForwardEstimator(
        freq="1H",
        prediction_length=prediction_length,
        context_length=context_length,
    ).create_training_network(device)
    delta = 2 * mean_boundary / num_groups
    dataset_group = []
    whole_data_list = []
    start = pd.Timestamp("01-01-2019", freq="1H")
    for gid in range(num_groups):
        parameter_mean = -mean_boundary + gid * delta
        # change the parameters of the model
        for p in net.parameters():
            p.data = torch.normal(parameter_mean, 0.1, size=p.data.shape)
        ts = torch.normal(0, 0.1, size=(1, context_length))
        for num_ts in range(num_time_steps):
            ts_slice = torch.Tensor(ts[0][-context_length:]).view(
                1, context_length
            )
            prediction = net.get_distr(ts_slice).sample((5000,))
            prediction = sum(prediction) / len(prediction)
            ts = torch.cat([ts, prediction], dim=1)
        whole_data_list.append(
            {
                "target": ts.view(
                    len(ts[0]),
                )[context_length:],
                "start": start,
            }
        )
        dataset_group.append(
            ListDataset(
                [
                    {
                        "target": ts.view(
                            len(ts[0]),
                        )[context_length:],
                        "start": start,
                    }
                ],
                freq="1H",
            )
        )
    dataset = ListDataset(whole_data_list, freq="1H")
    dataset_group = [dataset] + dataset_group

    # save to files
    with open("synthetic_whole_data.csv", "wb") as output:
        pickle.dump(dataset, output)

    with open("synthetic_group_data.csv", "wb") as output:
        pickle.dump(dataset_group, output)
    return True


def get_synthetic_data_mlp(
    model_name=None, num_groups=8, mean_boundary=0.5, num_duplicates=16
):
    assert num_groups > 1
    prediction_length = 1
    context_length = 12
    device = "cpu"
    dataset_group = []
    whole_data_list = []
    start = pd.Timestamp("01-01-2019", freq="1H")
    for gid in range(num_groups):
        net = SimpleFeedForwardEstimator(
            freq="1H",
            prediction_length=prediction_length,
            context_length=context_length,
        ).create_training_network(device)
        for p in net.parameters():
            p.data = torch.normal(0, 0.1, size=p.data.shape)
        pattern_group = []
        # for j in range(num_duplicates):
        # ts = torch.Uniform(0, 1, size=(1, context_length))
        while True:
            ts = torch.rand(size=(1, context_length))
            ts_slice = torch.Tensor(ts[0][-context_length:]).view(
                1, context_length
            )
            prediction = net.get_distr(ts_slice).sample((1000,))
            prediction = sum(prediction) / len(prediction)
            if abs(torch.norm(prediction)) <= 1:
                break
        ts = torch.cat([ts, prediction], dim=1)
        ts = ts.view(
            len(ts[0]),
        )  # [context_length:]
        for j in range(num_duplicates):
            ts_sample = ts + torch.normal(0, 0.1, size=ts.shape)
            whole_data_list.append({"target": ts_sample, "start": start})
            pattern_group.append({"target": ts_sample, "start": start})
        dataset_group.append(ListDataset(pattern_group, freq="1H"))
    random.shuffle(whole_data_list)
    random.shuffle(dataset_group)
    dataset = ListDataset(whole_data_list, freq="1H")
    ret = []
    ret.append(dataset)
    ret.append(dataset)
    dataset_group = [dataset] + dataset_group
    dataset_group = [dataset] + dataset_group

    # save to files
    with open("synthetic_mlp_whole_data.csv", "wb") as output:
        pickle.dump(ret, output)

    with open("synthetic_mlp_group_data.csv", "wb") as output:
        pickle.dump(dataset_group, output)
    return True


def get_synthetic_data_linear(
    context_length=24,
    prediction_length=8,
    num_groups=8,
    steps_per_ts=1,
    num_duplicates=16,
):
    assert num_groups > 1
    freq = "1H"
    len_sample = context_length + prediction_length

    dataset_group = []
    whole_data = []
    ret = dict()
    start = pd.Timestamp("01-01-2000", freq=freq)
    for gid in range(num_groups):
        model1 = torch.nn.Linear(context_length, prediction_length)
        model2 = torch.nn.Linear(context_length, prediction_length)
        # model1 = torch.sin
        # model2 = torch.cos
        pattern_group1 = []
        pattern_group2 = []
        sample_context = torch.rand(context_length)
        for t_step in range(2 * steps_per_ts):
            while True:
                with torch.no_grad():
                    if t_step <= steps_per_ts:
                        prediction = model1(sample_context)
                    else:
                        prediction = model2(sample_context)
                    if (
                        torch.norm(prediction) < prediction_length
                    ):  # and prediction_length*0.1 < torch.norm(prediction):
                        # prediction = torch.sin(prediction)
                        # prediction /= torch.max(prediction)
                        break
                    # prediction *= 10
                    # prediction += torch.normal(0, 0.1, size=prediction.shape)
            ts_sample = torch.cat([sample_context, prediction])
            # print(ts_sample)
            for j in range(num_duplicates):
                ts_sample += torch.normal(0, 0.1, size=ts_sample.shape)
                whole_data.append({"target": ts_sample, "start": start})
                if t_step <= steps_per_ts:
                    pattern_group1.append(
                        {"target": ts_sample, "start": start}
                    )
                else:
                    pattern_group2.append(
                        {"target": ts_sample, "start": start}
                    )
            sample_context = ts_sample[-context_length:]
            start += pd.Timedelta(hours=prediction_length)
        dataset_group.append(ListDataset(pattern_group1, freq=freq))
        """
        dataset_group.append(
            ListDataset(
                pattern_group2,
                freq=freq
            )
        )
        """
    print(len(whole_data))
    ret["group_ratio"] = [len(i) / len(whole_data) for i in dataset_group]
    print(ret["group_ratio"])
    random.shuffle(whole_data)
    ret["whole_data"] = ListDataset(whole_data, freq=freq)
    ret["group_data"] = dataset_group

    # save to files
    with open("synthetic_linear_new_whole_data.csv", "wb") as output:
        pickle.dump(ret["whole_data"], output)

    with open("synthetic_linear_new_group_data.csv", "wb") as output:
        pickle.dump(ret, output)

    return True


def get_synthetic_data_linear_simple(
    model_name=None, num_groups=8, mean_boundary=1, num_duplicates=16
):
    assert num_groups > 1
    num_time_steps = 100

    dataset_group = []
    whole_data_list = []
    start = pd.Timestamp("01-01-2019", freq="1D")
    for gid in range(num_groups):
        base = np.linspace(0, 10, num_time_steps)
        pattern_group = []
        ts = (
            (gid + 1)
            * mean_boundary
            * torch.FloatTensor(base).view(1, num_time_steps)
        )
        for j in range(num_duplicates):
            ts += torch.normal(0, 0.01, size=ts.shape)
            whole_data_list.append(
                {
                    "target": ts.view(
                        len(ts[0]),
                    ),
                    "start": start,
                }
            )
            pattern_group.append(
                {
                    "target": ts.view(
                        len(ts[0]),
                    ),
                    "start": start,
                }
            )
        dataset_group.append(ListDataset(pattern_group, freq="1D"))

    random.shuffle(whole_data_list)

    dataset = ListDataset(whole_data_list, freq="1D")
    dataset_group = [dataset] + dataset_group

    # save to files
    with open("synthetic_linear_simple_whole_data.csv", "wb") as output:
        pickle.dump(dataset, output)

    with open("synthetic_linear_simple_group_data.csv", "wb") as output:
        pickle.dump(dataset_group, output)
    return True


def get_synthetic_data_sin(
    model_name=None, num_groups=32, mean_boundary=1, num_duplicates=50
):
    assert num_groups > 1
    num_time_steps = 100

    dataset_group = []
    whole_data_list = []
    no_duplicate_whole_data_list = []
    start = pd.Timestamp("01-01-2019", freq="1D")
    for gid in range(num_groups):
        mean = (gid + 1) * mean_boundary
        base = np.linspace(0, mean, num_time_steps)
        pattern_group = []
        ts = (gid + 1) * torch.sin(torch.FloatTensor(base)).view(
            1, num_time_steps
        )
        ts += torch.FloatTensor((gid + 1) * base).view(1, num_time_steps)
        no_duplicate_whole_data_list.append(
            {
                "target": ts.view(
                    len(ts[0]),
                ),
                "start": start,
            }
        )
        for j in range(num_duplicates):
            ts += torch.normal(0, 0.1, size=ts.shape)
            whole_data_list.append(
                {
                    "target": ts.view(
                        len(ts[0]),
                    ),
                    "start": start,
                }
            )
            pattern_group.append(
                {
                    "target": ts.view(
                        len(ts[0]),
                    ),
                    "start": start,
                }
            )
        dataset_group.append(ListDataset(pattern_group, freq="1D"))

    random.shuffle(whole_data_list)
    random.shuffle(no_duplicate_whole_data_list)
    ret_whole_dataset = []
    dataset = ListDataset(whole_data_list, freq="1D")
    no_duplicate_dataset = ListDataset(no_duplicate_whole_data_list, freq="1D")
    ret_whole_dataset.append(dataset)
    ret_whole_dataset.append(dataset)
    dataset_group = [dataset] + dataset_group
    dataset_group = [dataset] + dataset_group

    # save to files
    with open("synthetic_complexsin_whole_data.csv", "wb") as output:
        pickle.dump(ret_whole_dataset, output)

    with open("synthetic_complexsin_group_data.csv", "wb") as output:
        pickle.dump(dataset_group, output)
    return True



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List

import torch
import torch.nn as nn


class ARNetworkBase(nn.Module):
    def __init__(
        self,
        prediction_length: int,
        context_length: int,
    ) -> None:
        super().__init__()

        self.prediction_length = prediction_length
        self.context_length = context_length
        # self.criterion = nn.MSELoss(reduction='none')
        self.criterion = nn.SmoothL1Loss(reduction="none")

        modules = []
        modules.append(nn.Linear(context_length, prediction_length))
        self.linear = nn.Sequential(*modules)


class ARTrainingNetwork(ARNetworkBase):
    def forward(
        self, past_target: torch.Tensor, future_target: torch.Tensor
    ) -> torch.Tensor:
        # m = max(torch.mean(past_target), torch.mean(future_target))
        # past_target /= m
        # future_target /= m
        nu = min(
            torch.mean(past_target).item(), torch.mean(future_target).item()
        )
        past_target /= 1 + nu
        future_target /= 1 + nu
        prediction = self.linear(past_target)
        loss = self.criterion(prediction, future_target)
        return loss


class ARPredictionNetwork(ARNetworkBase):
    def __init__(
        self, num_parallel_samples: int = 100, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.num_parallel_samples = num_parallel_samples

    def forward(self, past_target: torch.Tensor) -> torch.Tensor:
        pass



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from collections.abc import Sequence
from functools import singledispatch


def divide_into(length, n):
    base_size, r = divmod(length, n)

    partition_length = [base_size] * n

    for idx in range(r):
        partition_length[idx] += 1

    return partition_length


@singledispatch
def partition(xs, n):
    raise NotImplementedError(
        f"Type {type(xs)} has not implemeted `partition`."
    )


@partition.register(Sequence)
def partition_sequence(xs: Sequence, n):
    slices = divide_into(len(xs), n)

    start = 0
    partitions = []
    for slice_length in slices:
        partitions.append(xs[start : start + slice_length])
        start += slice_length

    return partitions



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
import json
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import cast, Dict, Tuple, Union
import numpy as np
import pandas as pd
from gluonts.dataset.common import Dataset, FileDataset, MetaData
from numpy import ma
from tsbench.constants import (
    DEFAULT_DATA_CATCH22_PATH,
    DEFAULT_DATA_PATH,
    DEFAULT_DATA_STATS_PATH,
)


@dataclass(frozen=True)
class DatasetConfig:
    """
    A dataset configuration references a dataset containing multiple time
    series.
    """

    base_path: Path = field(default=DEFAULT_DATA_PATH, compare=False)

    @classmethod
    def name(cls) -> str:
        """
        Returns a canonical name for the dataset.
        """
        raise NotImplementedError

    def generate(self) -> None:
        """
        Downloads the dataset into the globally configured data directory and
        applies necessary preprocessing steps.

        This function must be called on a machine prior to using the dataset.
        """
        raise NotImplementedError

    def prepare(self) -> None:
        """
        Generates all necessary representations of the dataset after it has
        been generated.
        """
        self.data.train().prepare()
        self.data.val().prepare()
        self.data.test().prepare()

    @property
    def has_time_features(self) -> bool:
        """
        Returns whether the dataset has time features.
        """
        return True

    @property
    def max_training_time(self) -> int:
        """
        Returns the maximum training time in seconds for the dataset.
        """
        raise NotImplementedError

    @property
    def meta(self) -> MetaData:
        """
        Returns the dataset's metadata.
        """
        return MetaData.parse_file(self.root / "gluonts" / "metadata.json")

    @property
    def data(self) -> DatasetSplits:
        """
        Returns the dataset's splits, i.e. training, validation, and test data.

        This is a noop, the datasets are only loaded at a later point.
        """
        return DatasetSplits(self.meta, self.root)

    @property
    def root(self) -> Path:
        """
        Returns the directory where all the data pertaining to this dataset is
        stored.
        """
        return self.base_path / self.name()

    def stats(
        self, root: str | Path = DEFAULT_DATA_STATS_PATH
    ) -> dict[str, float]:
        """
        Returns basic statistics of the dataset.

        Args:
            root: The directory which contains the stats files.
        """
        file = Path(root) / f"{self.name()}.json"
        with file.open("r") as f:
            return json.load(f)

    def catch22(
        self, root: str | Path = DEFAULT_DATA_CATCH22_PATH
    ) -> pd.DataFrame:
        """
        Returns the catch22 features of all time series in the dataset.

        Args:
            root: The directory which contains the catch22 feature files.
        """
        file = Path(root) / f"{self.name()}.parquet"
        return pd.read_parquet(file)


@dataclass
class DatasetSplits:
    """
    The dataset splits provide train, validation and test data for a particular
    dataset.

    Calling any of the functions here, is a noop. Data is only loaded once a
    particular representation of the data is accessed.
    """

    _metadata: MetaData
    _directory: Path

    def train(self, val: bool = True) -> DatasetSplit:
        """
        Returns the train data for the dataset.

        Args:
            val: Whether validation data is used. If not, this returns the validation data, i.e.
                the same time series that are longer by the prediction length.
        """
        return DatasetSplit(
            self._metadata, self._directory, "train" if val else "val"
        )

    def val(self) -> DatasetSplit:
        """
        Returns the validation data for the dataset.

        This is the same as :meth:`train(False)`.
        """
        return DatasetSplit(self._metadata, self._directory, "val")

    def test(self) -> DatasetSplit:
        """
        Returns the test data for the dataset.
        """
        return DatasetSplit(self._metadata, self._directory, "test")


@dataclass
class DatasetSplit:
    """
    A dataset split provides all the representations for a particular split
    (train/val/test) of a dataset.
    """

    _metadata: MetaData
    _directory: Path
    _split: str

    def gluonts(self) -> Dataset:
        """
        Returns the GluonTS dataset for the dataset split.

        This loads the associated JSON file and is, thus, potentially slow.
        """
        return FileDataset(
            self._directory / "gluonts" / self._split, freq=self._metadata.freq
        )

    def evaluation(self) -> EvaluationDataset:
        """
        Returns the NumPy arrays that are used to perform evaluation.
        """
        if self._split == "train":
            raise ValueError(
                "training data does not provide an evaluation dataset"
            )

        base = self._directory / "numpy" / self._split
        return EvaluationDataset(
            np.load(base / "future_data.npy"),
            ma.MaskedArray(
                np.load(base / "past_data.npy"),
                mask=np.load(base / "past_mask.npy"),
            ),
        )

    def prepare(self) -> None:
        """
        Prepares all required representations provided that the GluonTS dataset
        is already generated.
        """
        target = self._directory / "numpy" / self._split
        if self._split == "train":
            if target.exists():
                shutil.rmtree(target)
            return

        if target.exists():
            if (
                (target / "future_data.npy").exists()
                and (target / "past_data.npy").exists()
                and (target / "past_mask.npy").exists()
            ):
                return
            shutil.rmtree(target)

        target.mkdir(parents=True)
        future, past = _generate_evaluation_dataset(
            self.gluonts(), cast(int, self._metadata.prediction_length)
        )

        np.save(target / "future_data.npy", future)
        np.save(target / "past_data.npy", past.data)
        np.save(target / "past_mask.npy", past.mask)


@dataclass
class EvaluationDataset:
    """
    The evaluation dataset is a simple dataset representation that contains a
    two-dimensional array of future values as well as a two-dimensional
    (masked) array of the past values that a model sees during training.

    This representation is very efficient for evaluation.
    """

    future: np.ndarray
    past: ma.MaskedArray


# -------------------------------------------------------------------------------------------------


def _generate_evaluation_dataset(
    dataset: Dataset, prediction_length: int
) -> tuple[np.ndarray, ma.MaskedArray]:
    # Extract data from all the values in the dataset
    pasts = []
    past_lengths = []
    predictions = []
    for item in dataset:
        target = item["target"]
        pasts.append(target[:-prediction_length])
        past_lengths.append(target.shape[0] - prediction_length)
        predictions.append(target[-prediction_length:])

    # Compute masked past values as well as (non-masked) future values
    max_len = np.max(past_lengths)
    values = np.empty((len(dataset), max_len))
    mask = np.ones((len(dataset), max_len), dtype=bool)
    for i in range(len(dataset)):
        values[i, : past_lengths[i]] = pasts[i]
        mask[i, : past_lengths[i]] = False

    y_past = ma.masked_array(values, mask=mask)
    y_true = np.stack(predictions)
    return y_true, y_past



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import io

import pandas as pd
import numpy as np
import pytest

from gluonts.dataset import pandas


@pytest.fixture(params=[pd.date_range, pd.period_range])
def my_series(request):

    idx = request.param("2021-01-01", freq="1D", periods=3)
    series = pd.Series(np.random.normal(size=3), index=idx)
    return series


@pytest.fixture()
def my_dataframe(my_series):
    return my_series.to_frame(name="target")


# elements will be called with my_series fixture, doesn't include long-format
all_formats = [
    lambda series: series,
    lambda series: series.to_frame("target"),
    lambda series: 3 * [series],
    lambda series: 3 * [series.to_frame("target")],
    lambda series: {i: series for i in ["A", "B", "C"]},
    lambda series: {i: series.to_frame("target") for i in ["A", "B", "C"]},
]


@pytest.fixture()
def long_dataframe():
    N, T = 2, 10
    df = pd.DataFrame(index=np.arange(N * T))
    df["time"] = (
        2 * pd.date_range("2021-01-01 00:00", freq="1H", periods=T).to_list()
    )
    df["target"] = np.random.normal(size=(N * T,))
    df["item"] = T * ["A"] + T * ["B"]
    df["stat_cat_1"] = T * [0] + T * [1]
    df["dyn_real_1"] = np.random.normal(size=(N * T,))
    return df


@pytest.fixture
def long_dataset(long_dataframe):  # initialized with dict
    return pandas.PandasDataset.from_long_dataframe(
        dataframe=long_dataframe,
        target="target",
        timestamp="time",
        item_id="item",
        freq="1H",
        feat_dynamic_real=["dyn_real_1"],
        feat_static_cat=["stat_cat_1"],
    )


@pytest.mark.parametrize("get_data", all_formats)
def test_PandasDataset_init_with_all_formats(get_data, my_series):
    dataset = pandas.PandasDataset(dataframes=get_data(my_series))
    assert len(dataset)
    for i in dataset:
        assert isinstance(i, dict)


def test_LongDataFrameDataset_iter(long_dataset):
    for i in long_dataset:
        assert isinstance(i, dict)
        assert "start" in i
        assert "target" in i
        assert "feat_dynamic_real" in i
        assert "feat_static_cat" in i


def test_LongDataFrameDataset_len(long_dataset):
    assert len(long_dataset) == 2


def test_as_dataentry(long_dataframe):
    df = long_dataframe.groupby("item").get_group("A")
    dataentry = pandas.as_dataentry(
        data=df,
        target="target",
        timestamp="time",
        feat_dynamic_real=["dyn_real_1"],
        feat_static_cat=["stat_cat_1"],
    )
    assert "start" in dataentry
    assert "target" in dataentry
    assert "feat_dynamic_real" in dataentry
    assert "feat_static_cat" in dataentry


def test_prepare_prediction_data():
    assert np.all(
        pandas.prepare_prediction_data(
            {"target": np.arange(20)}, ignore_last_n_targets=5
        )["target"]
        == np.arange(15)
    )


def test_prepare_prediction_data_nested():
    assert np.all(
        pandas.prepare_prediction_data(
            {"target": np.ones(shape=(3, 20))},
            ignore_last_n_targets=5,
        )["target"]
        == np.ones(shape=(3, 15))
    )


def test_prepare_prediction_data_with_features():
    res = pandas.prepare_prediction_data(
        {
            "start": pd.Period("2021-01-01", freq="1H"),
            "target": np.array([1.0, 2.0, np.nan]),
            "feat_dynamic_real": np.array([[1.0, 2.0, 3.0]]),
            "past_feat_dynamic_real": np.array([[1.0, 2.0, np.nan]]),
        },
        ignore_last_n_targets=1,
    )
    expected = {
        "start": pd.Period("2021-01-01", freq="1H"),
        "target": np.array([1.0, 2.0]),
        "feat_dynamic_real": np.array([[1.0, 2.0, 3.0]]),
        "past_feat_dynamic_real": np.array([[1.0, 2.0]]),
    }
    for key in res:
        assert np.all(res[key] == expected[key])


def test_is_uniform_2H():
    timestamps = ["2021-01-01 00:00", "2021-01-01 02:00", "2021-01-01 04:00"]
    assert pandas.is_uniform(pd.DatetimeIndex(timestamps).to_period("2H"))


@pytest.mark.parametrize(
    "timestamps",
    [
        ["2021-01-01 00:00", "2021-01-01 02:00", "2021-01-01 02:00"],
        ["2021-01-01 00:00", "2021-01-01 02:00", "2021-01-01 03:00"],
        ["2021-01-01 04:00", "2021-01-01 02:00", "2021-01-01 00:00"],
    ],
)
def test_is_uniform_2H_fail(timestamps):
    assert not pandas.is_uniform(pd.DatetimeIndex(timestamps).to_period("2H"))


def test_infer_period(my_dataframe):
    ds = pandas.PandasDataset(my_dataframe, target="target", freq="1D")
    for entry in ds:
        assert entry["start"] == pd.Period("2021-01-01", freq="1D")


def test_infer_period2(my_dataframe):
    dfs = {"A": my_dataframe, "B": my_dataframe}
    ds = pandas.PandasDataset(dfs, target="target", freq="1D")
    for entry in ds:
        assert entry["start"] == pd.Period("2021-01-01", freq="1D")


def test_is_series(my_series):
    assert pandas.is_series(my_series)
    assert pandas.is_series([my_series])
    assert pandas.is_series({"A": my_series})


def test_is_series_fail(my_dataframe):
    with pytest.raises(AssertionError):
        assert pandas.is_series(my_dataframe)
    with pytest.raises(AssertionError):
        assert pandas.is_series([my_dataframe])
    with pytest.raises(AssertionError):
        assert pandas.is_series({"A": my_dataframe})


def test_series_to_dataframe(my_series):
    assert isinstance(pandas.series_to_dataframe(my_series), pd.DataFrame)
    assert isinstance(pandas.series_to_dataframe([my_series])[0], pd.DataFrame)
    dict_df = pandas.series_to_dataframe({"A": my_series})
    assert list(dict_df.keys())[0] == "A"
    assert isinstance(list(dict_df.values())[0], pd.DataFrame)


def test_long_csv_3M():
    data = (
        "timestamp,item_id,target\n"
        "2021-03,0,102\n"
        "2021-06,0,103\n"
        "2021-09,0,102\n"
        "2021-12,0,99\n"
        "2021-04,1,134\n"
        "2021-07,1,151\n"
        "2021-10,1,144\n"
        "2022-01,1,148\n"
        "2022-04,1,117\n"
        "2022-07,1,138\n"
        "2021-02,2,212\n"
        "2021-05,2,225\n"
        "2021-08,2,221\n"
        "2021-11,2,227\n"
        "2022-02,2,230\n"
        "2022-05,2,229\n"
    )

    with io.StringIO(data) as fp:
        ds = pandas.PandasDataset.from_long_dataframe(
            pd.read_csv(fp),
            target="target",
            item_id="item_id",
            timestamp="timestamp",
            freq="3M",
        )
        for entry in ds:
            assert entry["start"].freqstr == "3M"

    with io.StringIO(data) as fp:
        ds = pandas.PandasDataset.from_long_dataframe(
            pd.read_csv(fp, index_col="timestamp"),
            target="target",
            item_id="item_id",
            freq="3M",
        )
        for entry in ds:
            assert entry["start"].freqstr == "3M"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Train/test splitter
~~~~~~~~~~~~~~~~~~~

This module defines strategies to split a whole dataset into train and test
subsets.

For uniform datasets, where all time-series start and end at the same point in
time `OffsetSplitter` can be used::

    splitter = OffsetSplitter(prediction_length=24, split_offset=24)
    train, test = splitter.split(whole_dataset)

For all other datasets, the more flexible `DateSplitter` can be used::

    splitter = DateSplitter(
        prediction_length=24,
        split_date=pd.Period('2018-01-31', freq='D')
    )
    train, test = splitter.split(whole_dataset)

The module also supports rolling splits::

    splitter = DateSplitter(
        prediction_length=24,
        split_date=pd.Period('2018-01-31', freq='D')
    )
    train, test = splitter.rolling_split(whole_dataset, windows=7)
"""

from abc import ABC, abstractmethod
from typing import List, Optional, cast

import numpy as np
import pandas as pd
import pydantic

from gluonts.dataset.common import DataEntry
from gluonts.dataset.field_names import FieldName


class TimeSeriesSlice(pydantic.BaseModel):
    """
    Like DataEntry, but all time-related fields are of type pd.Series and is
    indexable, e.g `ts_slice['2018':]`.
    """

    class Config:
        arbitrary_types_allowed = True

    target: pd.Series
    item: str

    feat_static_cat: List[int] = []
    feat_static_real: List[float] = []

    feat_dynamic_cat: List[pd.Series] = []
    feat_dynamic_real: List[pd.Series] = []

    @classmethod
    def from_data_entry(
        cls, item: DataEntry, freq: Optional[str] = None
    ) -> "TimeSeriesSlice":
        if freq is None:
            freq = item["start"].freq

        index = pd.period_range(
            start=item["start"], freq=freq, periods=len(item["target"])
        )

        feat_dynamic_cat = [
            pd.Series(cat, index=index)
            for cat in list(item.get("feat_dynamic_cat", []))
        ]

        feat_dynamic_real = [
            pd.Series(real, index=index)
            for real in list(item.get("feat_dynamic_real", []))
        ]

        feat_static_cat = list(item.get("feat_static_cat", []))

        feat_static_real = list(item.get("feat_static_real", []))

        return TimeSeriesSlice(
            target=pd.Series(item["target"], index=index),
            item=item[FieldName.ITEM_ID],
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            feat_dynamic_cat=feat_dynamic_cat,
            feat_dynamic_real=feat_dynamic_real,
        )

    def to_data_entry(self) -> DataEntry:
        ret = {
            FieldName.START: self.start,
            FieldName.ITEM_ID: self.item,
            FieldName.TARGET: self.target.values,
        }

        if self.feat_static_cat:
            ret[FieldName.FEAT_STATIC_CAT] = self.feat_static_cat
        if self.feat_static_real:
            ret[FieldName.FEAT_STATIC_REAL] = self.feat_static_real
        if self.feat_dynamic_cat:
            ret[FieldName.FEAT_DYNAMIC_CAT] = [
                cast(np.ndarray, cat.values).tolist()
                for cat in self.feat_dynamic_cat
            ]
        if self.feat_dynamic_real:
            ret[FieldName.FEAT_DYNAMIC_REAL] = [
                cast(np.ndarray, real.values).tolist()
                for real in self.feat_dynamic_real
            ]

        return ret

    @property
    def start(self) -> pd.Period:
        return self.target.index[0]

    @property
    def end(self) -> pd.Period:
        return self.target.index[-1]

    def __len__(self) -> int:
        return len(self.target)

    def __getitem__(self, slice_: slice) -> "TimeSeriesSlice":
        feat_dynamic_real = []
        feat_dynamic_cat = []

        if self.feat_dynamic_real is not None:
            feat_dynamic_real = [
                feat[slice_] for feat in self.feat_dynamic_real
            ]

        if self.feat_dynamic_cat is not None:
            feat_dynamic_cat = [feat[slice_] for feat in self.feat_dynamic_cat]

        target = self.target[slice_]

        assert all([len(target) == len(feat) for feat in feat_dynamic_real])
        assert all([len(target) == len(feat) for feat in feat_dynamic_cat])

        return TimeSeriesSlice(
            target=target,
            item=self.item,
            feat_dynamic_cat=feat_dynamic_cat,
            feat_dynamic_real=feat_dynamic_real,
            feat_static_cat=self.feat_static_cat,
            feat_static_real=self.feat_static_real,
        )


class TrainTestSplit(pydantic.BaseModel):
    train: List[DataEntry] = []
    test: List[DataEntry] = []

    def _add_train_slice(self, train_slice: TimeSeriesSlice) -> None:
        # is there any data left for training?
        if train_slice:
            self.train.append(train_slice.to_data_entry())

    def _add_test_slice(self, test_slice: TimeSeriesSlice) -> None:
        self.test.append(test_slice.to_data_entry())


class AbstractBaseSplitter(ABC):
    """
    Base class for all other splitter.

    Args:
        param prediction_length:
            The prediction length which is used to train themodel.

        max_history:
            If given, all entries in the *test*-set have a max-length of
            `max_history`. This can be used to produce smaller file-sizes.
    """

    @abstractmethod
    def _train_slice(self, item: TimeSeriesSlice) -> TimeSeriesSlice:
        pass

    @abstractmethod
    def _test_slice(
        self, item: TimeSeriesSlice, offset: int = 0
    ) -> TimeSeriesSlice:
        pass

    def _trim_history(self, item: TimeSeriesSlice) -> TimeSeriesSlice:
        if getattr(self, "max_history") is not None:
            return item[-getattr(self, "max_history") :]
        else:
            return item

    def split(self, items: List[DataEntry]) -> TrainTestSplit:
        split = TrainTestSplit()

        for item in map(TimeSeriesSlice.from_data_entry, items):

            train = self._train_slice(item)
            test = self._trim_history(self._test_slice(item))

            split._add_train_slice(train)

            prediction_length = getattr(self, "prediction_length")

            _check_split_length(train.end, test.end, prediction_length)
            split._add_test_slice(test)

        return split

    def rolling_split(
        self,
        items: List[DataEntry],
        windows: int,
        distance: Optional[int] = None,
    ) -> TrainTestSplit:
        # distance defaults to prediction_length
        if distance is None:
            distance = getattr(self, "prediction_length")
        assert distance is not None

        split = TrainTestSplit()

        for item in map(TimeSeriesSlice.from_data_entry, items):
            train = self._train_slice(item)
            split._add_train_slice(train)

            for window in range(windows):
                offset = window * distance
                test = self._trim_history(
                    self._test_slice(item, offset=offset)
                )
                prediction_length = getattr(self, "prediction_length")

                _check_split_length(train.end, test.end, prediction_length)
                split._add_test_slice(test)

        return split


def _check_split_length(
    train_end: pd.Period, test_end: pd.Period, prediction_length: int
) -> None:
    msg = (
        "Not enough observations after the split point to construct"
        " the test instance; consider using longer time series,"
        " or splitting at an earlier point."
    )
    assert train_end + prediction_length <= test_end, msg


class OffsetSplitter(pydantic.BaseModel, AbstractBaseSplitter):
    """
    A splitter that slices training and test data based on a fixed integer
    offset.

    Parameters
    ----------
    prediction_length
        Length of the prediction interval in test data.
    split_offset
        Offset determining where the training data ends.
        A positive offset indicates how many observations since the start of
        each series should be in the training slice; a negative offset
        indicates how many observations before the end of each series should
        be excluded from the training slice. Please make sure that the number
        of excluded values is enough for the test case, i.e., at least
        ``prediction_length`` (for ``rolling_split`` multiple of
        ``prediction_length``) values are left off.
    max_history
        If given, all entries in the *test*-set have a max-length of
        `max_history`. This can be used to produce smaller file-sizes.
    """

    prediction_length: int
    split_offset: int
    max_history: Optional[int] = None

    def _train_slice(self, item: TimeSeriesSlice) -> TimeSeriesSlice:
        return item[: self.split_offset]

    def _test_slice(
        self, item: TimeSeriesSlice, offset: int = 0
    ) -> TimeSeriesSlice:
        offset_ = self.split_offset + offset + self.prediction_length
        if self.split_offset < 0 and offset_ >= 0:
            offset_ += len(item)
        return item[:offset_]


class DateSplitter(AbstractBaseSplitter, pydantic.BaseModel):
    """
    A splitter that slices training and test data based on a
    ``pandas.Period``.

    Training entries obtained from this class will be limited to observations
    up to (including) the given ``split_date``.

    Parameters
    ----------
    prediction_length
        Length of the prediction interval in test data.
    split_date
        Period determining where the training data ends. Please make sure
        at least ``prediction_length`` (for ``rolling_split`` multiple of
        ``prediction_length``) values are left over after the ``split_date``.
    max_history
        If given, all entries in the *test*-set have a max-length of
        `max_history`. This can be used to produce smaller file-sizes.
    """

    class Config:
        arbitrary_types_allowed = True

    prediction_length: int
    split_date: pd.Period
    max_history: Optional[int] = None

    def _train_slice(self, item: TimeSeriesSlice) -> TimeSeriesSlice:
        # the train-slice includes everything up to (including) the split date
        return item[: self.split_date]

    def _test_slice(
        self, item: TimeSeriesSlice, offset: int = 0
    ) -> TimeSeriesSlice:
        return item[
            : self.split_date
            + (self.prediction_length + offset) * item.start.freq
        ]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional

import torch
import torch.nn as nn

from pts import Trainer
from pts.dataset import FieldName
from pts.model import PTSEstimator, PTSPredictor, copy_parameters
from pts.modules import DistributionOutput, StudentTOutput
from pts.transform import (
    Transformation,
    Chain,
    InstanceSplitter,
    ExpectedNumInstanceSampler,
)
from pts.transform.sampler import CustomUniformSampler
from .ar_network import (
    ARTrainingNetwork,
    ARPredictionNetwork,
)


class AREstimator(PTSEstimator):
    def __init__(
        self,
        prediction_length: int,
        freq: int,
        trainer: Trainer = Trainer(),
        context_length: Optional[int] = None,
        num_parallel_samples: int = 100,
    ) -> None:
        """
        Defines an estimator.

        All parameters should be serializable.
        """
        super().__init__(trainer=trainer)
        self.num_parallel_samples = num_parallel_samples
        self.freq = freq

        self.prediction_length = prediction_length
        self.context_length = (
            context_length if context_length is not None else prediction_length
        )

    # here we do only a simple operation to convert the input data to a form
    # that can be digested by our model by only splitting the target in two, a
    # conditioning part and a to-predict part, for each training example.
    # For a more complex transformation example, see the `pts.model.deepar`
    # transformation that includes time features, age feature, observed values
    # indicator, etc.
    def create_transformation(self, is_full_batch=False) -> Transformation:
        return Chain(
            [
                InstanceSplitter(
                    target_field=FieldName.TARGET,
                    is_pad_field=FieldName.IS_PAD,
                    start_field=FieldName.START,
                    forecast_start_field=FieldName.FORECAST_START,
                    # train_sampler=ExpectedNumInstanceSampler(num_instances=1),
                    train_sampler=CustomUniformSampler(),
                    past_length=self.context_length,
                    future_length=self.prediction_length,
                    is_full_batch=is_full_batch,
                    time_series_fields=[],  # [FieldName.FEAT_DYNAMIC_REAL]
                )
            ]
        )

    # defines the network, we get to see one batch to initialize it.
    # the network should return at least one tensor that is used as a loss to minimize in the training loop.
    # several tensors can be returned for instance for analysis, see DeepARTrainingNetwork for an example.
    def create_training_network(
        self, device: torch.device
    ) -> ARTrainingNetwork:
        return ARTrainingNetwork(
            prediction_length=self.prediction_length,
            context_length=self.context_length,
        ).to(device)

    # we now define how the prediction happens given that we are provided a
    # training network.
    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: nn.Module,
        device: torch.device,
    ) -> PTSPredictor:
        prediction_network = ARPredictionNetwork(
            prediction_length=self.prediction_length,
            context_length=self.context_length,
            num_parallel_samples=self.num_parallel_samples,
        ).to(device)

        copy_parameters(trained_network, prediction_network)

        return PTSPredictor(
            input_transform=transformation,
            prediction_net=prediction_network,
            batch_size=self.trainer.batch_size,
            freq=self.freq,
            prediction_length=self.prediction_length,
            device=device,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
import sys
from distutils.util import strtobool
from typing import ContextManager

import numpy as np
import pytest

from gluonts.core.component import equals
from gluonts.dataset.jsonl import encode_json
from gluonts.model.trivial.mean import MeanPredictor
from gluonts.shell.env import ServeEnv, TrainEnv
from gluonts.shell.train import run_train_and_test

try:
    from gluonts.shell.serve import Settings
    from gluonts.testutil import shell as testutil
except ImportError:
    if sys.platform != "win32":
        raise

    # gunicorn doesn't work on windows, so we skip these tests
    pytestmark = pytest.mark.skip


context_length = 5
prediction_length = 6
num_samples = 4


@pytest.fixture(scope="function")  # type: ignore
def train_env(listify_dataset) -> ContextManager[TrainEnv]:
    hyperparameters = {
        "context_length": context_length,
        "prediction_length": prediction_length,
        "num_samples": num_samples,
        "listify_dataset": listify_dataset,
        "num_workers": 3,
        "num_prefetch": 4,
        "shuffle_buffer_length": 256,
        "epochs": 3,
        "quantiles": [0.1, 0.25, 0.5, 0.75, 0.9],
        "test_quantiles": [0.25, 0.75],
    }
    with testutil.temporary_train_env(hyperparameters, "constant") as env:
        yield env


@pytest.fixture(scope="function")  # type: ignore
def static_server(
    train_env: TrainEnv,
) -> ContextManager["testutil.ServerFacade"]:
    predictor = MeanPredictor.from_hyperparameters(**train_env.hyperparameters)
    predictor.serialize(train_env.path.model)

    serve_env = ServeEnv(train_env.path.base)
    settings = Settings(sagemaker_server_port=testutil.free_port())
    with testutil.temporary_server(serve_env, None, settings) as server:
        yield server


@pytest.fixture(scope="function")  # type: ignore
def dynamic_server(
    train_env: TrainEnv,
) -> ContextManager["testutil.ServerFacade"]:
    serve_env = ServeEnv(train_env.path.base)
    settings = Settings(sagemaker_server_port=testutil.free_port())
    with testutil.temporary_server(
        serve_env, MeanPredictor, settings
    ) as server:
        yield server


@pytest.fixture
def batch_transform(monkeypatch, train_env):
    monkeypatch.setenv("SAGEMAKER_BATCH", "true")

    inference_config = {
        "context_length": context_length,
        "prediction_length": prediction_length,
        "num_samples": num_samples,
        "output_types": ["mean", "samples"],
        "quantiles": [],
        **train_env.hyperparameters,
    }

    monkeypatch.setenv("INFERENCE_CONFIG", json.dumps(inference_config))
    monkeypatch.setenv("GLUONTS_FORWARD_FIELDS", json.dumps(["foo"]))
    return inference_config


@pytest.mark.parametrize("listify_dataset", ["yes", "no"])
def test_listify_dataset(train_env: TrainEnv, listify_dataset):
    for dataset in train_env.datasets.values():
        if listify_dataset == "yes":
            assert isinstance(dataset, list)


@pytest.mark.parametrize("listify_dataset", ["yes", "no"])
@pytest.mark.parametrize("forecaster_type", [MeanPredictor])
def test_train_shell(train_env: TrainEnv, caplog, forecaster_type) -> None:
    run_train_and_test(env=train_env, forecaster_type=forecaster_type)

    if forecaster_type == MeanPredictor:
        for _, _, line in caplog.record_tuples:
            if "#test_score (local, QuantileLoss" in line:
                assert line.endswith("0.0")
            if "local, wQuantileLoss" in line:
                assert line.endswith("0.0")
            if "local, Coverage" in line:
                assert line.endswith("0.0")
            if "MASE" in line or "MSIS" in line:
                assert line.endswith("nan")
            if "abs_target_sum" in line:
                assert line.endswith("270.0")


@pytest.mark.parametrize("listify_dataset", ["yes", "no"])
def test_server_shell(
    train_env: TrainEnv, static_server: "testutil.ServerFacade", caplog
) -> None:
    execution_parameters = static_server.execution_parameters()

    assert "BatchStrategy" in execution_parameters
    assert "MaxConcurrentTransforms" in execution_parameters
    assert "MaxPayloadInMB" in execution_parameters

    assert execution_parameters["BatchStrategy"] == "SINGLE_RECORD"
    assert execution_parameters["MaxPayloadInMB"] == 6

    configuration = {
        "num_samples": 1,  # FIXME: this is ignored
        "output_types": ["mean", "samples"],
        "quantiles": [],
        **train_env.hyperparameters,
    }

    for entry in train_env.datasets["train"]:
        forecast = static_server.invocations([entry], configuration)[0]

        for output_type in configuration["output_types"]:
            assert output_type in forecast

        act_mean = np.array(forecast["mean"])
        act_samples = np.array(forecast["samples"])

        mean = np.mean(entry["target"])

        exp_mean_shape = (prediction_length,)
        exp_samples_shape = (num_samples, prediction_length)

        exp_mean = mean * np.ones(shape=(prediction_length,))
        exp_samples = mean * np.ones(shape=exp_samples_shape)

        assert exp_mean_shape == act_mean.shape
        assert exp_samples_shape == act_samples.shape
        assert equals(exp_mean, act_mean)
        assert equals(exp_samples, act_samples)


@pytest.mark.parametrize("listify_dataset", ["yes", "no"])
def test_dynamic_shell(
    train_env: TrainEnv, dynamic_server: "testutil.ServerFacade", caplog
) -> None:
    execution_parameters = dynamic_server.execution_parameters()

    assert "BatchStrategy" in execution_parameters
    assert "MaxConcurrentTransforms" in execution_parameters
    assert "MaxPayloadInMB" in execution_parameters

    assert execution_parameters["BatchStrategy"] == "SINGLE_RECORD"
    assert execution_parameters["MaxPayloadInMB"] == 6

    configuration = {
        "num_eval_samples": 1,  # FIXME: this is ignored
        "output_types": ["mean", "samples"],
        "quantiles": [],
        **train_env.hyperparameters,
    }

    for entry in train_env.datasets["train"]:
        forecast = dynamic_server.invocations([entry], configuration)[0]

        for output_type in configuration["output_types"]:
            assert output_type in forecast

        act_mean = np.array(forecast["mean"])
        act_samples = np.array(forecast["samples"])

        mean = np.mean(entry["target"])

        exp_mean_shape = (prediction_length,)
        exp_samples_shape = (num_samples, prediction_length)

        exp_mean = mean * np.ones(shape=(prediction_length,))
        exp_samples = mean * np.ones(shape=exp_samples_shape)

        assert exp_mean_shape == act_mean.shape
        assert exp_samples_shape == act_samples.shape
        assert equals(exp_mean, act_mean)
        assert equals(exp_samples, act_samples)


@pytest.mark.parametrize("listify_dataset", ["yes", "no"])
def test_dynamic_batch_shell(
    batch_transform,
    train_env: TrainEnv,
    dynamic_server: "testutil.ServerFacade",
    caplog,
) -> None:
    execution_parameters = dynamic_server.execution_parameters()

    assert "BatchStrategy" in execution_parameters
    assert "MaxConcurrentTransforms" in execution_parameters
    assert "MaxPayloadInMB" in execution_parameters

    assert execution_parameters["BatchStrategy"] == "SINGLE_RECORD"
    assert execution_parameters["MaxPayloadInMB"] == 6

    for entry in train_env.datasets["train"]:
        entry["foo"] = 42
        forecast = dynamic_server.batch_invocations([entry])[0]

        for output_type in batch_transform["output_types"]:
            assert output_type in forecast

        assert forecast["foo"] == 42

        act_mean = np.array(forecast["mean"])
        act_samples = np.array(forecast["samples"])

        mean = np.mean(entry["target"])

        exp_mean_shape = (prediction_length,)
        exp_samples_shape = (num_samples, prediction_length)

        exp_mean = mean * np.ones(shape=(prediction_length,))
        exp_samples = mean * np.ones(shape=exp_samples_shape)

        assert exp_mean_shape == act_mean.shape
        assert exp_samples_shape == act_samples.shape
        assert equals(exp_mean, act_mean)
        assert equals(exp_samples, act_samples)


def test_as_json_dict_outputs_valid_json():
    non_compliant_json = {
        "a": float("nan"),
        "k": float("infinity"),
        "b": {
            "c": float("nan"),
            "d": "testing",
            "e": float("-infinity"),
            "f": float("infinity"),
            "g": {"h": float("nan")},
        },
    }

    with pytest.raises(ValueError):
        json.dumps(non_compliant_json, allow_nan=False)

    json.dumps(encode_json(non_compliant_json), allow_nan=False)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
This example shows how to fit a model and evaluate its predictions.
"""
import pprint

from gluonts.dataset.repository.datasets import get_dataset, dataset_recipes
from gluonts.evaluation import make_evaluation_predictions, Evaluator
from gluonts.model.simple_feedforward import SimpleFeedForwardEstimator
from gluonts.mx.trainer import Trainer


if __name__ == "__main__":

    print(f"datasets available: {dataset_recipes.keys()}")

    # we pick m4_hourly as it only contains a few hundred time series
    dataset = get_dataset("m4_hourly", regenerate=False)

    estimator = SimpleFeedForwardEstimator(
        prediction_length=dataset.metadata.prediction_length,
        trainer=Trainer(epochs=5, num_batches_per_epoch=10),
    )

    predictor = estimator.train(dataset.train)

    forecast_it, ts_it = make_evaluation_predictions(
        dataset.test, predictor=predictor, num_samples=100
    )

    agg_metrics, item_metrics = Evaluator()(
        ts_it, forecast_it, num_series=len(dataset.test)
    )

    pprint.pprint(agg_metrics)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.
from typing import Optional

import pandas as pd

from .field_names import FieldName


def forecast_start(entry, time_axis: int = -1):
    return entry[FieldName.START] + entry[FieldName.TARGET].shape[time_axis]


def to_pandas(instance: dict, freq: Optional[str] = None) -> pd.Series:
    """
    Transform a dictionary into a pandas.Series object, using its "start" and
    "target" fields.

    Parameters
    ----------
    instance
        Dictionary containing the time series data.
    freq
        Frequency to use in the pandas.Series index.

    Returns
    -------
    pandas.Series
        Pandas time series object.
    """
    target = instance[FieldName.TARGET]
    start = instance[FieldName.START]
    if not freq:
        freq = start.freqstr

    return pd.Series(
        target,
        index=pd.period_range(start=start, periods=len(target), freq=freq),
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Iterator

import mxnet as mx
from gluonts.model.forecast_generator import (
    make_distribution_forecast,
    recursively_zip_arrays,
)
from gluonts.mx.distribution import Distribution
from gluonts.mx.model.forecast import DistributionForecast


@recursively_zip_arrays.register(mx.nd.NDArray)
def _(x: mx.nd.NDArray) -> Iterator[mx.nd.NDArray]:
    for i in range(x.shape[0]):
        yield x[i]


@make_distribution_forecast.register(Distribution)
def _(distr: Distribution, *args, **kwargs) -> DistributionForecast:
    return DistributionForecast(distr, *args, **kwargs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# pylint: disable=missing-function-docstring
import math
import numpy as np
import numpy.typing as npt
from numpy import ma
from .quantile import QuantileForecasts


def rmse(
    y_pred: npt.NDArray[np.float32], y_true: npt.NDArray[np.float32]
) -> float:
    return np.sqrt(((y_pred - y_true) ** 2).mean())


def abs_error_sum(
    y_pred: npt.NDArray[np.float32], y_true: npt.NDArray[np.float32]
) -> float:
    return np.abs(y_pred - y_true).sum()


def abs_target_sum(y_true: npt.NDArray[np.float32]) -> float:
    return np.sum(np.abs(y_true))


def abs_target_mean(y_true: npt.NDArray[np.float32]) -> float:
    return np.mean(np.abs(y_true))


def naive_error(
    y_past: ma.masked_array,  # type: ignore
    seasonality: int,
) -> npt.NDArray[np.float32]:
    error = np.abs(y_past[:, seasonality:] - y_past[:, :-seasonality]).mean(1)
    return ma.getdata(error)


def mase(
    y_pred: npt.NDArray[np.float32],
    y_true: npt.NDArray[np.float32],
    error: npt.NDArray[np.float32],
) -> float:
    mase_values = np.abs(y_pred - y_true).mean(1) / error
    return mase_values.mean()


def smape(
    y_pred: npt.NDArray[np.float32], y_true: npt.NDArray[np.float32]
) -> float:
    median = y_pred
    num = np.abs(y_true - median)
    denom = (np.abs(y_true) + np.abs(median)) / 2
    # If the denominator is 0, we set it to float('inf') such that any division yields 0 (this
    # might not be fully mathematically correct, but at least we don't get NaNs)
    denom[denom == 0] = math.inf
    return np.mean(num / denom, axis=1).mean()


def ncrps(y_pred: QuantileForecasts, y_true: npt.NDArray[np.float32]) -> float:
    y_true_rep = y_true[:, None].repeat(len(y_pred.quantiles), axis=1)
    quantiles = np.array([float(q) for q in y_pred.quantiles])
    quantile_losses = 2 * np.sum(
        np.abs(
            (y_pred.values - y_true_rep)
            * ((y_true_rep <= y_pred.values) - quantiles[:, None])
        ),
        axis=-1,
    )  # shape [num_time_series, num_quantiles]
    denom = np.sum(np.abs(y_true))  # shape [1]
    weighted_losses = quantile_losses.sum(0) / denom  # shape [num_quantiles]
    return weighted_losses.mean()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging
import re
from typing import Dict, Iterator, NamedTuple, Optional, Tuple

import pandas as pd

import gluonts  # noqa
from gluonts.core.serde import load_code
from gluonts.dataset.common import DataEntry, Dataset
from gluonts.dataset.field_names import FieldName
from gluonts.dataset.stat import (
    DatasetStatistics,
    calculate_dataset_statistics,
)
from gluonts.evaluation import Evaluator
from gluonts.model.estimator import Estimator
from gluonts.model.forecast import Forecast
from gluonts.model.predictor import Predictor
from gluonts.itertools import maybe_len
from gluonts.transform import AdhocTransform


def make_evaluation_predictions(
    dataset: Dataset,
    predictor: Predictor,
    num_samples: int = 100,
) -> Tuple[Iterator[Forecast], Iterator[pd.Series]]:
    """
    Returns predictions for the trailing prediction_length observations of the
    given time series, using the given predictor.

    The predictor will take as input the given time series without the trailing
    prediction_length observations.

    Parameters
    ----------
    dataset
        Dataset where the evaluation will happen. Only the portion excluding
        the prediction_length portion is used when making prediction.
    predictor
        Model used to draw predictions.
    num_samples
        Number of samples to draw on the model when evaluating. Only
        sampling-based models will use this.

    Returns
    -------
    Tuple[Iterator[Forecast], Iterator[pd.Series]]
        A pair of iterators, the first one yielding the forecasts, and the
        second one yielding the corresponding ground truth series.
    """

    prediction_length = predictor.prediction_length
    lead_time = predictor.lead_time

    def add_ts_dataframe(
        data_iterator: Iterator[DataEntry],
    ) -> Iterator[DataEntry]:
        for data_entry in data_iterator:
            data = data_entry.copy()
            index = pd.period_range(
                start=data[FieldName.START],
                periods=data[FieldName.TARGET].shape[-1],
                freq=data[FieldName.START].freq,
            )
            data["ts"] = pd.DataFrame(
                index=index, data=data[FieldName.TARGET].transpose()
            )
            yield data

    def ts_iter(dataset: Dataset) -> pd.DataFrame:
        for data_entry in add_ts_dataframe(iter(dataset)):
            yield data_entry["ts"]

    def truncate_target(data):
        data = data.copy()
        target = data[FieldName.TARGET]
        assert (
            target.shape[-1] >= prediction_length
        )  # handles multivariate case (target_dim, history_length)
        data[FieldName.TARGET] = target[..., : -prediction_length - lead_time]
        return data

    # TODO filter out time series with target shorter than prediction length
    # TODO or fix the evaluator so it supports missing values instead (all
    # TODO the test set may be gone otherwise with such a filtering)

    dataset_trunc = AdhocTransform(truncate_target).apply(dataset)

    return (
        predictor.predict(dataset_trunc, num_samples=num_samples),
        ts_iter(dataset),
    )


train_dataset_stats_key = "train_dataset_stats"
test_dataset_stats_key = "test_dataset_stats"
estimator_key = "estimator"
agg_metrics_key = "agg_metrics"


def serialize_message(logger, message: str, variable):
    logger.info(f"gluonts[{message}]: {variable}")


def backtest_metrics(
    test_dataset: Dataset,
    predictor: Predictor,
    evaluator=Evaluator(
        quantiles=(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9)
    ),
    num_samples: int = 100,
    logging_file: Optional[str] = None,
) -> Tuple[dict, pd.DataFrame]:
    """
    Parameters
    ----------
    test_dataset
        Dataset to use for testing.
    predictor
        The predictor to test.
    evaluator
        Evaluator to use.
    num_samples
        Number of samples to use when generating sample-based forecasts. Only
        sampling-based models will use this.
    logging_file
        If specified, information of the backtest is redirected to this file.

    Returns
    -------
    Tuple[dict, pd.DataFrame]
        A tuple of aggregate metrics and per-time-series metrics obtained by
        training `forecaster` on `train_dataset` and evaluating the resulting
        `evaluator` provided on the `test_dataset`.
    """

    if logging_file is not None:
        log_formatter = logging.Formatter(
            "[%(asctime)s %(levelname)s %(thread)d] %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
        )
        logger = logging.getLogger(__name__)
        handler = logging.FileHandler(logging_file)
        handler.setFormatter(log_formatter)
        logger.addHandler(handler)
    else:
        logger = logging.getLogger(__name__)

    test_statistics = calculate_dataset_statistics(test_dataset)
    serialize_message(logger, test_dataset_stats_key, test_statistics)

    forecast_it, ts_it = make_evaluation_predictions(
        test_dataset, predictor=predictor, num_samples=num_samples
    )

    agg_metrics, item_metrics = evaluator(
        ts_it, forecast_it, num_series=maybe_len(test_dataset)
    )

    # we only log aggregate metrics for now as item metrics may be very large
    for name, value in agg_metrics.items():
        serialize_message(logger, f"metric-{name}", value)

    if logging_file is not None:
        # Close the file handler to avoid letting the file open.
        # https://stackoverflow.com/questions/24816456/python-logging-wont-shutdown
        logger.removeHandler(handler)
        del logger, handler

    return agg_metrics, item_metrics


# TODO does it make sense to have this then?
class BacktestInformation(NamedTuple):
    train_dataset_stats: DatasetStatistics
    test_dataset_stats: DatasetStatistics
    estimator: Estimator
    agg_metrics: Dict[str, float]

    @staticmethod
    def make_from_log(log_file):
        with open(log_file) as f:
            return BacktestInformation.make_from_log_contents(
                "\n".join(f.readlines())
            )

    @staticmethod
    def make_from_log_contents(log_contents):
        messages = dict(re.findall(r"gluonts\[(.*)\]: (.*)", log_contents))

        # avoid to fail if a key is missing for instance in the case a run did
        # not finish so that we can still get partial information
        try:
            return BacktestInformation(
                train_dataset_stats=eval(
                    messages[train_dataset_stats_key]
                ),  # TODO: use load
                test_dataset_stats=eval(
                    messages[test_dataset_stats_key]
                ),  # TODO: use load
                estimator=load_code(messages[estimator_key]),
                agg_metrics={
                    k: load_code(v)
                    for k, v in messages.items()
                    if k.startswith("metric-") and v != "nan"
                },
            )
        except Exception as error:
            logging.error(error)
            return None



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Iterator, List, Tuple, TypeVar
from sklearn.model_selection import LeaveOneGroupOut
from tsbench.config import Config, EnsembleConfig, ModelConfig
from tsbench.evaluations.metrics import Performance
from tsbench.evaluations.tracking import Tracker

T = TypeVar("T", ModelConfig, EnsembleConfig)


def loocv_split(
    tracker: Tracker[T],
) -> Iterator[
    Tuple[
        Tuple[List[Config[T]], List[Performance]],
        Tuple[List[Config[T]], List[Performance]],
    ]
]:
    """
    Iterates over the configurations and associated performances obtained from
    a collector. For each item it yields, it leaves out
    configurations/performances for a single dataset for testing and provides
    the configurations/performances for training.

    Args:
        tracker: The tracker to retrieve data from.
        show_progerss: Whether to show progress via tqdm.

    Yields:
        The training data, including configurations and performances, and the test data, including
        configurations and performances.
    """
    data = tracker.get_evaluations()
    groups = [c.dataset.name() for c in data.configurations]

    # Split the data according to the datasets
    loocv = LeaveOneGroupOut()
    for I_train, I_test in loocv.split(data.configurations, groups=groups):
        X_train = [data.configurations[i] for i in I_train]
        y_train = [data.performances[i] for i in I_train]

        X_test = [data.configurations[i] for i in I_test]
        y_test = [data.performances[i] for i in I_test]

        yield (X_train, y_train), (X_test, y_test)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
import tarfile
import tempfile
from functools import partial
from pathlib import Path
from typing import Any, cast, Dict, List, Optional
import botocore
import click
from tqdm.auto import tqdm
from tqdm.contrib.concurrent import process_map
from tsbench.analysis.utils import run_parallel
from tsbench.constants import DEFAULT_EVALUATIONS_PATH
from tsbench.evaluations import aws
from tsbench.evaluations.aws import default_session
from tsbench.evaluations.tracking.job import Job, load_jobs_from_analysis
from ._main import evaluations


@evaluations.command(short_help="Download evaluations to your file system.")
@click.option(
    "--experiment",
    type=str,
    default=None,
    help=(
        "The AWS Sagemaker experiment from which to download the evaluations."
        " If not provided, downloads the publicly available evaluations"
        " (s3://odp-tsbench)."
    ),
)
@click.option(
    "--include_forecasts",
    type=bool,
    default=False,
    help=(
        "Whether to download forecasts (plenty of data) or "
        "only the training, validation and testing metrics."
    ),
)
@click.option(
    "--evaluations_path",
    type=click.Path(),
    default=DEFAULT_EVALUATIONS_PATH,
    show_default=True,
    help="The path to which to download the evaluations to.",
)
def download(
    experiment: Optional[str], include_forecasts: bool, evaluations_path: str
):
    """
    Downloads either the evaluations of a single AWS Sagemaker experiment or
    the publicly available evaluations.

    The evaluations are downloaded to the provided directory.
    """
    target = Path(evaluations_path)
    target.mkdir(parents=True, exist_ok=True)

    if experiment is None:
        print("Downloading publicly available evaluations...")
        _download_public_evaluations(
            include_forecasts=include_forecasts, evaluations_path=target
        )
    else:
        print(f"Downloading data from experiment '{experiment}'...")
        analysis = aws.Analysis(experiment)
        process_map(
            partial(
                _move_job, target=target, include_forecasts=include_forecasts
            ),
            load_jobs_from_analysis(analysis),
            chunksize=1,
        )


def _download_public_evaluations(
    include_forecasts: bool, evaluations_path: Path
) -> None:
    public_bucket = "odp-tsbench"
    session = default_session()
    client = session.client(
        "s3",
        config=botocore.client.Config(  # type: ignore
            signature_version=botocore.UNSIGNED,
            max_pool_connections=2 * cast(int, os.cpu_count()),
        ),
    )

    # First, download the metrics
    print("Downloading metrics...")
    with tempfile.TemporaryDirectory() as tmp:
        file = Path(tmp) / "metrics.tar.gz"
        client.download_file(public_bucket, "metrics.tar.gz", str(file))
        with tarfile.open(file, mode="r:gz") as tar:
            tar.extractall(evaluations_path)

    # Then, optionally download the forecasts
    if include_forecasts:
        print("Downloading forecasts...")

        # First, get all files
        with tqdm(desc="List objects") as progress:
            response = client.list_objects(Bucket=public_bucket)
            objects = _extract_object_names(response)
            progress.update()
            while response["IsTruncated"]:
                response = client.list_objects(
                    Bucket=public_bucket, Marker=objects[-1]
                )
                objects.extend(_extract_object_names(response))
                progress.update()

        # Then, download all of the objects
        run_parallel(
            partial(
                _download_object,
                bucket=public_bucket,
                client=client,
                destination=evaluations_path,
            ),
            objects,
            2 * cast(int, os.cpu_count()),
        )


def _download_object(
    key: str, bucket: str, client: Any, destination: Path
) -> None:
    target = destination / key
    target.parent.mkdir(parents=True, exist_ok=True)
    client.download_file(Bucket=bucket, Key=key, Filename=str(target))


def _extract_object_names(response: Dict[str, Any]) -> List[str]:
    return [
        obj["Key"]
        for obj in response["Contents"]
        if not obj["Key"].endswith("/")
        and obj["Key"] != "metrics.tar.gz"
        and not obj["Key"].endswith("config.json")
        and not obj["Key"].endswith("performance.json")
    ]


def _move_job(job: Job, target: Path, include_forecasts: bool):
    job.save(target, include_forecasts=include_forecasts)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pickle
import random
from functools import partial
from itertools import combinations, product
from pathlib import Path
from typing import cast, List, Tuple
import click
from tsbench.analysis import EnsembleAnalyzer
from tsbench.analysis.utils import num_fitting_processes, run_parallel
from tsbench.config import DatasetConfig, ModelConfig, TrainConfig
from tsbench.constants import DEFAULT_EVALUATIONS_PATH
from tsbench.evaluations.metrics import Performance
from tsbench.evaluations.tracking import ModelTracker
from ._main import ensembles


@ensembles.command(short_help="Simulate the performance of ensembles.")
@click.option(
    "--output_path",
    type=click.Path(),
    default=Path.home() / "ensembles" / "evaluation.pickle",
    show_default=True,
    help="The local file where evaluated ensembles should be written to.",
)
@click.option(
    "--data_path",
    type=click.Path(exists=True),
    default=Path.home() / "data" / "datasets",
    show_default=True,
    help="The local path where the datasets are stored.",
)
@click.option(
    "--evaluations_path",
    type=click.Path(exists=True),
    default=DEFAULT_EVALUATIONS_PATH,
    show_default="True",
    help="The local path where offline evaluations can be found.",
)
@click.option(
    "--max_ensemble_size",
    default=10,
    help=(
        "The maximum number of ensemble members for any ensemble that is"
        " evaluated."
    ),
)
@click.option(
    "--default_samples",
    default=0,
    show_default=True,
    help=(
        "The number of ensembles to sample which combine only default"
        " configurations of all models. If set to -1, uses all available"
        " configurations."
    ),
)
@click.option(
    "--hyperensemble_samples",
    default=0,
    show_default=True,
    help=(
        "The number of ensembles to sample which constitute hyperensembles. If"
        " set to -1, usesall hyperensembles that can be built."
    ),
)
@click.option(
    "--random_samples",
    default=0,
    show_default=True,
    help=(
        "The number of ensembles to sample which combine random configurations"
        " found in theoffline evaluations."
    ),
)
@click.option(
    "--sample_datasets",
    default=True,
    show_default=True,
    help=(
        "Whether to sample a single random dataset for each ensemble being"
        " evaluated or toevaluate each ensemble on all available datasets."
    ),
)
@click.option(
    "--seed",
    default=0,
    show_default=True,
    help="Seed for reproducible sampling.",
)
def simulate(
    data_path: str,
    evaluations_path: str,
    output_path: str,
    max_ensemble_size: int,
    default_samples: int,
    hyperensemble_samples: int,
    random_samples: int,
    sample_datasets: bool,
    seed: int,
):
    """
    Simulates the performance of various ensembles.

    The ensembles are built from configurations (i.e. model types and
    hyperparameters) for which offline evaluations are available.
    """
    assert any(
        [default_samples != 0, hyperensemble_samples != 0, random_samples != 0]
    ), "No samples are specified."

    random.seed(seed)

    # Load the experiments
    print("Loading experiments...")
    tracker = ModelTracker.from_directory(
        Path(evaluations_path), data_path=Path(data_path)
    )

    # Sample configurations
    print("Sampling configurations...")
    unique_configurations = tracker.unique_model_configs()
    default_configurations = [
        cast(ModelConfig, c)
        for c in unique_configurations
        if not isinstance(c, TrainConfig) or c == c.__class__()
    ]
    choices: List[Tuple[ModelConfig]] = []

    # If desired, we combine all base configurations into ensembles of sizes between 2 and the
    # provided maximum. Then, we potentially sample from this collection. For 13 default
    # configurations and a maximum ensemble size of 10, this results in 8,086 ensembles.
    if default_samples != 0:
        available_ensembles = [
            combination
            for i in range(2, max_ensemble_size + 1)
            for combination in combinations(default_configurations, i)
        ]
        if default_samples == -1:
            choices.extend(available_ensembles)
        else:
            choices.extend(random.sample(available_ensembles, default_samples))

    if hyperensemble_samples != 0:
        available_ensembles = []
        for config in default_configurations:
            all_configs = [
                c
                for c in unique_configurations
                if isinstance(c, config.__class__)
                and (
                    not isinstance(c, TrainConfig) or c.training_fraction == 1
                )
            ]
            if len(all_configs) == 1:
                continue
            hyper_ensembles = [
                combination
                for i in range(2, max_ensemble_size + 1)
                for combination in combinations(all_configs, i)
            ]
            available_ensembles.extend(hyper_ensembles)

        if hyperensemble_samples == -1:
            choices.extend(available_ensembles)
        else:
            choices.extend(
                random.sample(available_ensembles, hyperensemble_samples)
            )

    # Then, we add some randomly sampled ensembles of model configurations
    for _ in range(random_samples):
        ensemble_size = random.randrange(2, max_ensemble_size + 1)
        configs = random.sample(unique_configurations, ensemble_size)
        choices.append(tuple(configs))

    # Then, we either evaluate each chosen configuration on all datasets or on a randomly sampled
    # one.
    datasets = list(
        {c.dataset for c in tracker.get_evaluations().configurations}
    )
    if sample_datasets:
        evaluations = [
            (model_config, random.choice(datasets)) for model_config in choices
        ]
    else:
        evaluations = list(product(choices, datasets))

    # Eventually, we can evaluate the ensembles that we have sampled
    print("Evaluating ensembles...")
    evaluator = EnsembleAnalyzer(tracker)
    results = run_parallel(
        partial(_evaluate_ensemble, evaluator=evaluator),
        evaluations,
        num_fitting_processes(cpus_per_process=1, memory_per_process=8),
    )

    # Afterwards, we can store all configurations along with their results. For now, we are just
    # storing them as pickled objects.
    with Path(output_path).open("wb+") as f:
        pickle.dump(
            [
                {
                    "configurations": list(evaluation[0]),
                    "dataset": evaluation[1],
                    "performance": result,
                }
                for evaluation, result in zip(evaluations, results)
            ],
            f,
        )


def _evaluate_ensemble(
    ensemble: Tuple[Tuple[ModelConfig], DatasetConfig],
    evaluator: EnsembleAnalyzer,
) -> Performance:
    return evaluator.get_ensemble_performance(
        list(ensemble[0]), ensemble[1], num_samples=3
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.
from pprint import pprint

import pandas as pd

from gluonts.dataset.repository.datasets import get_dataset
from gluonts.model.simple_feedforward import SimpleFeedForwardEstimator
from gluonts.mx.trainer import Trainer
from gluonts.evaluation import make_evaluation_predictions, Evaluator
from gluonts.dataset.rolling_dataset import (
    StepStrategy,
    generate_rolling_dataset,
)

if __name__ == "__main__":
    dataset = get_dataset("constant", regenerate=False)

    estimator = SimpleFeedForwardEstimator(
        prediction_length=dataset.metadata.prediction_length,
        trainer=Trainer(epochs=5, num_batches_per_epoch=10),
    )

    predictor = estimator.train(dataset.train)

    # create the rolled dataset to use for forecasting and evaluation
    dataset_rolled = generate_rolling_dataset(
        dataset=dataset.test,
        start_time=pd.Timestamp("2000-01-01-15", freq="1H"),
        end_time=pd.Timestamp("2000-01-02-04", freq="1H"),
        strategy=StepStrategy(
            prediction_length=dataset.metadata.prediction_length,
        ),
    )

    forecast_it, ts_it = make_evaluation_predictions(
        dataset_rolled, predictor=predictor, num_samples=len(dataset_rolled)
    )

    agg_metrics, _ = Evaluator()(ts_it, forecast_it)

    pprint(agg_metrics)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path
import click
from tsbench.constants import DEFAULT_EVALUATIONS_PATH
from tsbench.utils import compress_directory
from ._main import evaluations


@evaluations.command(
    short_help="Archive metrics of all evaluations into a single file."
)
@click.option(
    "--evaluations_path",
    type=click.Path(exists=True),
    default=DEFAULT_EVALUATIONS_PATH,
    help="The directory where TSBench evaluations are stored.",
)
@click.option(
    "--archive_path",
    type=click.Path(),
    default=Path.home() / "archive",
    help="The directory to which to write the compressed files.",
)
def archive(evaluations_path: str, archive_path: str):
    """
    Archives the metrics of all evaluations found in the provided directory
    into a single file.

    This is most probably only necessary for publishing the metrics in an
    easier format.
    """
    source = Path(evaluations_path)
    target = Path(archive_path)
    target.mkdir(parents=True, exist_ok=True)

    # First, tar all the metadata
    print("Compressing metrics...")
    compress_directory(
        source,
        target / "metrics.tar.gz",
        include={"config.json", "performance.json"},
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pts.dataset.repository.datasets import get_dataset, dataset_recipes
from pts.dataset.utils import to_pandas
import numpy as np
import pandas as pd
from pts.dataset import ListDataset
import torch
from pts.model.simple_feedforward import SimpleFeedForwardEstimator
import pickle
import json
import random


def group_electricity_cv(
    num_ts=10,
    num_groups=14,
    context_length=72,
    prediction_length=12,
    file_name="default",
):
    dataset = get_dataset("electricity")
    len_sample = context_length + prediction_length
    dataset_group = [[] for i in range(num_groups)]
    train_full_data = []
    test_full_data = []
    ret = dict()
    train_it = iter(dataset.train)
    test_it = iter(dataset.test)
    # num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    date_checkpoint = [
        "2012-03-01",
        "2012-06-01",
        "2012-09-01",
        "2012-12-01",
        "2013-03-01",
        "2013-06-01",
        "2013-09-01",
        "2013-12-01",
        "2014-03-01",
    ]
    # get ready the training data
    for i in range(num_ts):
        train_entry = next(train_it)
        unsplit_ts = train_entry["target"]
        unsplit_start = train_entry["start"]
        t = unsplit_start
        start_date = 4

        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            for j, date_ckpt in enumerate(date_checkpoint):
                if unsplit_start < pd.Timestamp(date_ckpt):
                    sid = j
                    break
                elif unsplit_start > pd.Timestamp(date_checkpoint[-1]):
                    sid = len(date_checkpoint)
                    break
            gid = ((start_date + 1) % 7) + sid * 7
            start_date += 1
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            train_full_data.append(
                # {'target': ts_slice, 'start': t, 'feat_static_cat': train_entry['feat_static_cat']}
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": np.array([gid]),
                }
            )
            dataset_group[gid].append(
                # {'target': ts_slice, 'start': t, 'feat_static_cat': train_entry['feat_static_cat']}
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": np.array([gid]),
                }
            )
            unsplit_start += pd.Timedelta(hours=prediction_length)

    # get ready the test data
    for i in range(int(num_ts * 0.2)):
        test_entry = next(test_it)
        unsplit_ts = test_entry["target"]
        unsplit_start = test_entry["start"]
        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            test_full_data.append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": test_entry["feat_static_cat"],
                }
            )

    print("total number of training examples: ", len(train_full_data))
    ret["group_ratio"] = [len(i) / len(train_full_data) for i in dataset_group]
    print("ratio for each group: ", ret["group_ratio"])
    random.shuffle(train_full_data)
    ret["whole_data"] = ListDataset(
        train_full_data, freq=dataset.metadata.freq
    )
    random.shuffle(test_full_data)
    ret["val_data"] = ListDataset(test_full_data, freq=dataset.metadata.freq)
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq=dataset.metadata.freq))
    ret["group_data"] = group_data_list
    with open("../dataset/" + file_name + ".csv", "wb") as output:
        pickle.dump(ret, output)
    return True


def group_electricity_mb(
    num_ts=10,
    num_groups=14,
    context_length=72,
    prediction_length=12,
    file_name="default",
):
    dataset = get_dataset("traffic")
    len_sample = context_length + prediction_length
    dataset_group = [[] for i in range(num_groups)]
    train_full_data = []
    test_full_data = []
    ret = dict()
    train_it = iter(dataset.train)
    test_it = iter(dataset.test)
    # num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    date_checkpoint = ["2016-01-01"]
    # get ready the training data
    for i in range(num_ts):
        train_entry = next(train_it)
        unsplit_ts = train_entry["target"]
        unsplit_start = train_entry["start"]
        t = unsplit_start
        start_date = 4

        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            for j, date_ckpt in enumerate(date_checkpoint):
                if unsplit_start < pd.Timestamp(date_ckpt):
                    sid = j
                    break
                elif unsplit_start > pd.Timestamp(date_checkpoint[-1]):
                    sid = len(date_checkpoint)
                    break
            gid = ((start_date + 1) % 7) + sid * 7
            start_date += 1
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            train_full_data.append(
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            dataset_group[gid].append(
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            unsplit_start += pd.Timedelta(hours=prediction_length)

    # get ready the test data
    for i in range(int(num_ts * 0.2)):
        test_entry = next(test_it)
        unsplit_ts = test_entry["target"]
        unsplit_start = test_entry["start"]
        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            test_full_data.append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": test_entry["feat_static_cat"],
                }
            )

    print("total number of training examples: ", len(train_full_data))
    ret["group_ratio"] = [len(i) / len(train_full_data) for i in dataset_group]
    print("ratio for each group: ", ret["group_ratio"])
    random.shuffle(train_full_data)
    ret["whole_data"] = ListDataset(
        train_full_data, freq=dataset.metadata.freq
    )
    random.shuffle(test_full_data)
    ret["val_data"] = ListDataset(test_full_data, freq=dataset.metadata.freq)
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq=dataset.metadata.freq))
    ret["group_data"] = group_data_list
    with open("../dataset/" + file_name + ".csv", "wb") as output:
        pickle.dump(ret, output)
    return True



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
from dataclasses import dataclass
from typing import cast, Dict, List, Union
import numpy as np
import pandas as pd
from .metric import Metric


@dataclass
class Performance:
    """
    The performance class encapsulates the metrics that are recorded for
    configurations.
    """

    training_time: Metric
    latency: Metric
    num_model_parameters: Metric
    num_gradient_updates: Metric

    ncrps: Metric
    mase: Metric
    smape: Metric
    nrmse: Metric
    nd: Metric

    @classmethod
    def from_dict(cls, metrics: dict[str, float | int]) -> Performance:
        """
        Initializes a new performance object from the given 1D dictionary.

        Metrics are expected to be provided via `<metric>_mean` and
        `<metric>_std` keys.
        """
        kwargs = {
            m: Metric(metrics[f"{m}_mean"], metrics[f"{m}_std"])
            for m in cls.metrics()
        }
        return Performance(**kwargs)  # type: ignore

    @classmethod
    def metrics(cls) -> list[str]:
        """
        Returns the list of metrics that are exposed by the performance class.
        """
        # pylint: disable=no-member
        return list(cls.__dataclass_fields__.keys())  # type: ignore

    @classmethod
    def to_dataframe(
        cls, performances: list[Performance], std: bool = True
    ) -> pd.DataFrame:
        """
        Returns a data frame representing the provided performances.
        """
        fields = sorted(
            Performance.__dataclass_fields__.keys()
        )  # pylint: disable=no-member
        result = np.empty((len(performances), 18 if std else 9))

        offset = 2 if std else 1
        for i, performance in enumerate(performances):
            for j, field in enumerate(fields):
                result[i, j * offset] = cast(
                    Metric, getattr(performance, field)
                ).mean
                if std:
                    result[i, j * offset + 1] = cast(
                        Metric, getattr(performance, field)
                    ).std

        return pd.DataFrame(
            result,
            columns=[
                f
                for field in fields
                for f in (
                    [f"{field}_mean", f"{field}_std"]
                    if std
                    else [f"{field}_mean"]
                )
            ],
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import argparse
import json
import logging
import os
from pathlib import Path

from gluonts.core import serde
from gluonts.dataset import common
from gluonts.dataset.repository import datasets
from gluonts.evaluation import Evaluator, backtest

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(name)s %(message)s",
    datefmt="[%Y-%m-%d %H:%M:%S]",
)
logger = logging.getLogger(__name__)

# TODO: implement model_fn, input_fn, predict_fn, and output_fn !!
# TODO: segment script for readability


def train(arguments):
    """
    Generic train method that trains a specified estimator on a specified
    dataset.
    """

    logger.info("Downloading estimator config.")
    estimator_config = Path(arguments.estimator) / "estimator.json"
    with estimator_config.open() as config_file:
        estimator = serde.load_json(config_file.read())

    logger.info("Downloading dataset.")
    if arguments.s3_dataset is None:
        # load built in dataset
        dataset = datasets.get_dataset(arguments.dataset)
    else:
        # load custom dataset
        s3_dataset_dir = Path(arguments.s3_dataset)
        dataset = common.load_datasets(
            metadata=s3_dataset_dir,
            train=s3_dataset_dir / "train",
            test=s3_dataset_dir / "test",
        )

    logger.info("Starting model training.")
    predictor = estimator.train(dataset.train)
    forecast_it, ts_it = backtest.make_evaluation_predictions(
        dataset=dataset.test,
        predictor=predictor,
        num_samples=int(arguments.num_samples),
    )

    logger.info("Starting model evaluation.")
    evaluator = Evaluator(quantiles=eval(arguments.quantiles))

    agg_metrics, item_metrics = evaluator(
        ts_it, forecast_it, num_series=len(list(dataset.test))
    )

    # required for metric tracking.
    for name, value in agg_metrics.items():
        logger.info(f"gluonts[metric-{name}]: {value}")

    # save the evaluation results
    metrics_output_dir = Path(arguments.output_data_dir)
    with open(metrics_output_dir / "agg_metrics.json", "w") as f:
        json.dump(agg_metrics, f)
    with open(metrics_output_dir / "item_metrics.csv", "w") as f:
        item_metrics.to_csv(f, index=False)

    # save the model
    model_output_dir = Path(arguments.model_dir)
    predictor.serialize(model_output_dir)


if __name__ == "__main__":
    # TODO switch to click
    parser = argparse.ArgumentParser()

    # an alternative way to load hyperparameters via SM_HPS environment variable.
    parser.add_argument(
        "--sm-hps", type=json.loads, default=os.environ["SM_HPS"]
    )

    # input data, output dir and model directories
    parser.add_argument(
        "--model-dir", type=str, default=os.environ["SM_MODEL_DIR"]
    )
    parser.add_argument(
        "--output-data_dir", type=str, default=os.environ["SM_OUTPUT_DATA_DIR"]
    )

    parser.add_argument(
        "--input-dir", type=str, default=os.environ["SM_INPUT_DIR"]
    )

    parser.add_argument(
        "--estimator", type=str, default=os.environ["SM_CHANNEL_ESTIMATOR"]
    )
    # argument possibly not set
    parser.add_argument(
        "--s3-dataset",
        type=str,
        default=os.environ.get("SM_CHANNEL_S3_DATASET"),
    )
    parser.add_argument(
        "--dataset", type=str, default=os.environ["SM_HP_DATASET"]
    )
    parser.add_argument(
        "--num-samples", type=int, default=os.environ["SM_HP_NUM_SAMPLES"]
    )
    parser.add_argument(
        "--quantiles", type=str, default=os.environ["SM_HP_QUANTILES"]
    )

    args, _ = parser.parse_known_args()

    train(args)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
from mxnet.gluon import nn
from .base import Callback


class ParameterCountCallback(Callback):  # type: ignore
    """
    This callback allows counting model parameters during training.

    Attributes:
        num_parameters: The number of parameters of the model. This attribute should only be
            accessed after training.
    """

    def __init__(self) -> None:
        self.num_parameters = 0

    def on_network_initialization_end(self, network: nn.HybridBlock) -> None:
        self.num_parameters = sum(
            np.prod(p.shape) for p in network.collect_params().values()
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# pylint: disable=missing-function-docstring
import logging
import tempfile
from pathlib import Path
from typing import Any, Dict
import pytorch_lightning as pl
from sacred import Experiment
from tsbench.analysis import SurrogateAnalyzer
from tsbench.constants import DEFAULT_DATA_PATH, DEFAULT_EVALUATIONS_PATH
from tsbench.evaluations.tracking import ModelTracker
from tsbench.surrogate import create_surrogate

ex = Experiment()


@ex.config
def experiment_config():
    # pylint: disable=unused-variable
    experiment = "test"  # type: ignore
    data_path = str(DEFAULT_DATA_PATH)  # type: ignore
    evaluations_path = str(DEFAULT_EVALUATIONS_PATH)  # type: ignore
    surrogate = "nonparametric"  # type: ignore

    metrics = "ncrps_mean,latency_mean"  # type: ignore
    inputs = {  # type: ignore
        "use_simple_dataset_features": False,
        "use_seasonal_naive_performance": False,
        "use_catch22_features": False,
    }
    outputs = {  # type: ignore
        "normalization": "quantile",
        "imputation": False,
    }

    xgboost = {  # type: ignore
        "objective": "regression",
    }
    autogluon = {  # type: ignore
        "time_limit": 10,
    }
    mlp = {  # type: ignore
        "objective": "ranking",
        "discount": "linear",
        "hidden_layer_sizes": [32, 32],
        "weight_decay": 0.01,
        "dropout": 0.0,
    }


@ex.automain
def main(
    _config: Dict[str, Any],
    _seed: int,
    data_path: str,
    evaluations_path: str,
    surrogate: str,
    metrics: str,
    inputs: Dict[str, bool],
    outputs: Dict[str, Any],
):
    print(_config)

    pl.seed_everything(_seed)
    logging.getLogger("pytorch_lightning").setLevel(logging.WARNING)

    # First, get the tracker
    print("Fetching the data...")
    tracker = ModelTracker.from_directory(
        Path(evaluations_path), data_path=Path(data_path)
    )

    # Then, initialize the surrogate
    print("Initializing the surrogate...")
    metrics_list = metrics.split(",")
    surrogate_instance = create_surrogate(
        surrogate,
        tracker=tracker,
        predict=metrics_list,
        input_flags=inputs,
        output_normalization=outputs["normalization"],
        impute_simulatable=outputs["imputation"],
        **(_config[surrogate] if surrogate in _config else {})
    )

    # And evaluate it
    print("Evaluating the surrogate...")
    evaluator = SurrogateAnalyzer(
        surrogate_instance, tracker=tracker, metrics=metrics_list
    )
    result = evaluator.run()

    print(result.mean())

    # Eventually, we store the results
    print("Storing the results...")
    with tempfile.TemporaryDirectory() as d:
        path = Path(d) / "results.parquet"
        result.to_parquet(str(path))
        ex.add_artifact(path, content_type="application/octet-stream")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import logging
import os
import random
import sys
import warnings
from pathlib import Path

import numpy as np
import pytest

import gluonts

try:
    import mxnet as mx
except ImportError:
    mx = None


@pytest.fixture(scope="function", autouse=True)
def function_scope_seed(request):
    """A function scope fixture that manages rng seeds.
    This fixture automatically initializes the python, numpy and mxnet random
    number generators randomly on every test run.
    def test_ok_with_random_data():
        ...
    To fix the seed used for a test case mark the test function with the
    desired seed:
    @pytest.mark.seed(1)
    def test_not_ok_with_random_data():
        '''This testcase actually works.'''
        assert 17 == random.randint(0, 100)
    When a test fails, the fixture outputs the seed used. The user can then set
    the environment variable MXNET_TEST_SEED to the value reported, then rerun
    the test with:
        pytest --verbose -s <test_module_name.py> -k <failing_test>
    To run a test repeatedly, install pytest-repeat and add the --count argument:
        pip install pytest-repeat
        pytest --verbose -s <test_module_name.py> -k <failing_test> --count 1000
    """

    seed = request.node.get_closest_marker("seed")
    env_seed_str = os.getenv("MXNET_TEST_SEED")

    if seed is not None:
        seed = seed.args[0]
        assert isinstance(seed, int)
    elif env_seed_str is not None:
        seed = int(env_seed_str)
    else:
        seed = int.from_bytes(os.urandom(4), "big")

    post_test_state = np.random.get_state()
    np.random.seed(seed)
    random.seed(seed)

    if mx is not None:
        mx.random.seed(seed)

    seed_message = (
        "np/mx/python random seeds are set to "
        f"{seed}, use MXNET_TEST_SEED={seed} to reproduce."
    )

    # Always log seed on DEBUG log level. This makes sure we can find out the
    # value of the seed even if the test case causes a segfault and subsequent
    # teardown code is not run.
    logging.debug(seed_message)

    yield  # run the test

    np.random.set_state(post_test_state)


@pytest.fixture(autouse=True)
def doctest(doctest_namespace):
    doctest_namespace["np"] = np
    doctest_namespace["gluonts"] = gluonts

    if mx is not None:
        doctest_namespace["mx"] = mx
        doctest_namespace["gluon"] = mx.gluon

    import doctest

    doctest.ELLIPSIS_MARKER = "-etc-"


def get_collect_ignores():
    test_folder = Path(__file__).parent.resolve()

    old_path = sys.path
    sys.path = [
        path for path in sys.path if Path(path).resolve() != test_folder
    ]

    excludes = []

    for path in test_folder.glob("**/require-packages.txt"):
        with path.open() as requirements:
            for requirement in map(str.strip, requirements):
                try:
                    __import__(requirement)
                except ImportError:
                    excludes.append(str(path.parent.relative_to(test_folder)))
                    break

    if excludes:
        warnings.warn(
            f"Skipping tests because some packages are not installed: {excludes}"
        )

    sys.path = old_path
    return excludes


collect_ignore = get_collect_ignores()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# pylint: disable=too-many-lines
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Any, cast, Dict, List, Tuple
import numpy as np
import pandas as pd
from ._factory import register_dataset
from .preprocessing import ConstantTargetFilter, Filter
from .sources import (
    GluonTsDatasetConfig,
    KaggleDatasetConfig,
    M3DatasetConfig,
    MonashDatasetConfig,
)


@register_dataset
@dataclass(frozen=True)
class ExchangeRateDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `exchange_rate_nips` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "exchange_rate"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _gluonts_name(self) -> str:
        return "exchange_rate_nips"


@register_dataset
@dataclass(frozen=True)
class ElectricityDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `electricity_nips` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "electricity"

    @property
    def max_training_time(self) -> int:
        return 14400


@register_dataset
@dataclass(frozen=True)
class SolarDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `solar_nips` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "solar"

    @property
    def max_training_time(self) -> int:
        return 7200

    @property
    def _gluonts_name(self) -> str:
        return "solar_nips"


@register_dataset
@dataclass(frozen=True)
class WikiDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `wiki-rolling_nips` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "wiki"

    @property
    def max_training_time(self) -> int:
        return 14400

    @property
    def _gluonts_name(self) -> str:
        return "wiki-rolling_nips"


@register_dataset
@dataclass(frozen=True)
class TaxiDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `taxi_30min` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "taxi"

    @property
    def max_training_time(self) -> int:
        return 14400

    @property
    def _gluonts_name(self) -> str:
        return "taxi_30min"


@register_dataset
@dataclass(frozen=True)
class M3MonthlyDatasetConfig(M3DatasetConfig):
    """
    The dataset configuration for the `m3_monthly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m3_monthly"

    @property
    def max_training_time(self) -> int:
        return 7200


@register_dataset
@dataclass(frozen=True)
class M3QuarterlyDatasetConfig(M3DatasetConfig):
    """
    The dataset configuration for the `m3_quarterly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m3_quarterly"

    @property
    def max_training_time(self) -> int:
        return 3600


@register_dataset
@dataclass(frozen=True)
class M3YearlyDatasetConfig(M3DatasetConfig):
    """
    The dataset configuration for the `m3_yearly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m3_yearly"

    @property
    def max_training_time(self) -> int:
        return 3600


@register_dataset
@dataclass(frozen=True)
class M3OtherDatasetConfig(M3DatasetConfig):
    """
    The dataset configuration for the `m3_other` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m3_other"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def has_time_features(self) -> bool:
        return False


@register_dataset
@dataclass(frozen=True)
class M4HourlyDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `m4_hourly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m4_hourly"

    @property
    def max_training_time(self) -> int:
        return 7200

    @property
    def has_time_features(self) -> bool:
        return False


@register_dataset
@dataclass(frozen=True)
class M4DailyDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `m4_daily` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m4_daily"

    @property
    def max_training_time(self) -> int:
        return 14400

    @property
    def has_time_features(self) -> bool:
        return False


@register_dataset
@dataclass(frozen=True)
class M4WeeklyDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `m4_weekly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m4_weekly"

    @property
    def max_training_time(self) -> int:
        return 7200

    @property
    def has_time_features(self) -> bool:
        return False


@register_dataset
@dataclass(frozen=True)
class M4MonthlyDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `m4_monthly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m4_monthly"

    @property
    def max_training_time(self) -> int:
        return 28800

    @property
    def has_time_features(self) -> bool:
        return False


@register_dataset
@dataclass(frozen=True)
class M4QuarterlyDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `m4_quarterly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m4_quarterly"

    @property
    def max_training_time(self) -> int:
        return 14400

    @property
    def has_time_features(self) -> bool:
        return False


@register_dataset
@dataclass(frozen=True)
class M4YearlyDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `m4_yearly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m4_yearly"

    @property
    def max_training_time(self) -> int:
        return 7200

    @property
    def has_time_features(self) -> bool:
        return False


@register_dataset
@dataclass(frozen=True)
class M5DatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `m5` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "m5"

    @property
    def max_training_time(self) -> int:
        return 28800

    def _materialize(self, directory: Path, regenerate: bool = False) -> None:
        shutil.copytree(
            Path.home() / ".mxnet" / "gluon-ts" / "datasets" / "m5",
            directory / "m5",
        )
        super()._materialize(directory, regenerate=True)


@register_dataset
@dataclass(frozen=True)
class TourismMonthlyDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `tourism_monthly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "tourism_monthly"

    @property
    def max_training_time(self) -> int:
        return 7200


@register_dataset
@dataclass(frozen=True)
class TourismQuarterlyDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `tourism_quarterly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "tourism_quarterly"

    @property
    def max_training_time(self) -> int:
        return 3600


@register_dataset
@dataclass(frozen=True)
class TourismYearlyDatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `tourism_yearly` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "tourism_yearly"

    @property
    def max_training_time(self) -> int:
        return 3600


@register_dataset
@dataclass(frozen=True)
class NN5DatasetConfig(GluonTsDatasetConfig):
    """
    The dataset configuration for the `nn5_daily_without_missing` dataset.
    """

    @classmethod
    def name(cls) -> str:
        return "nn5"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _gluonts_name(self) -> str:
        return "nn5_daily_without_missing"


@register_dataset
@dataclass(frozen=True)
class LondonSmartMetersDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "London Smart Meters".
    """

    @classmethod
    def name(cls) -> str:
        return "london_smart_meters"

    @property
    def max_training_time(self) -> int:
        return 28800

    @property
    def _file(self) -> str:
        return "london_smart_meters_dataset_without_missing_values.zip"

    @property
    def _record(self) -> str:
        return "4656091"

    @property
    def _prediction_length(self) -> int:
        return 48


@register_dataset
@dataclass(frozen=True)
class WindFarmsDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Wind Farms".
    """

    @classmethod
    def name(cls) -> str:
        return "wind_farms"

    @property
    def max_training_time(self) -> int:
        return 28800

    def _filters(self, prediction_length: int) -> List[Filter]:
        return [
            ConstantTargetFilter(prediction_length, required_length=100000)
        ]

    @property
    def _file(self) -> str:
        return "wind_farms_minutely_dataset_without_missing_values.zip"

    @property
    def _record(self) -> str:
        return "4654858"

    @property
    def _prediction_length(self) -> int:
        return 60


@register_dataset
@dataclass(frozen=True)
class CarPartsDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Car Parts".
    """

    @classmethod
    def name(cls) -> str:
        return "car_parts"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _file(self) -> str:
        return "car_parts_dataset_without_missing_values.zip"

    @property
    def _record(self) -> str:
        return "4656021"

    @property
    def _prediction_length(self) -> int:
        return 12


@register_dataset
@dataclass(frozen=True)
class DominickDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Dominick".
    """

    @classmethod
    def name(cls) -> str:
        return "dominick"

    @property
    def max_training_time(self) -> int:
        return 28800

    @property
    def has_time_features(self) -> bool:
        return False

    @property
    def _file(self) -> str:
        return "dominick_dataset.zip"

    @property
    def _record(self) -> str:
        return "4654802"

    @property
    def _prediction_length(self) -> int:
        return 8

    @property
    def _prediction_length_multiplier(self) -> int:
        return 1


@register_dataset
@dataclass(frozen=True)
class FredMdDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Federal Reserve Economic Dataset".
    """

    @classmethod
    def name(cls) -> str:
        return "fred_md"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _file(self) -> str:
        return "fred_md_dataset.zip"

    @property
    def _record(self) -> str:
        return "4654833"

    @property
    def _prediction_length(self) -> int:
        return 12


@register_dataset
@dataclass(frozen=True)
class SanFranciscoTrafficDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "San Francisco Traffic".
    """

    @classmethod
    def name(cls) -> str:
        return "san_francisco_traffic"

    @property
    def max_training_time(self) -> int:
        return 28800

    @property
    def _file(self) -> str:
        return "traffic_hourly_dataset.zip"

    @property
    def _record(self) -> str:
        return "4656132"

    @property
    def _prediction_length(self) -> int:
        return 48


@register_dataset
@dataclass(frozen=True)
class PedestrianCountDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Pedestrian Counts".
    """

    @classmethod
    def name(cls) -> str:
        return "pedestrian_count"

    @property
    def max_training_time(self) -> int:
        return 14400

    @property
    def _file(self) -> str:
        return "pedestrian_counts_dataset.zip"

    @property
    def _record(self) -> str:
        return "4656626"

    @property
    def _prediction_length(self) -> int:
        return 48


@register_dataset
@dataclass(frozen=True)
class HospitalDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Hospitals".
    """

    @classmethod
    def name(cls) -> str:
        return "hospital"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _file(self) -> str:
        return "hospital_dataset.zip"

    @property
    def _record(self) -> str:
        return "4656014"

    @property
    def _prediction_length(self) -> int:
        return 12


@register_dataset
@dataclass(frozen=True)
class CovidDeathsDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "COVID Deaths".
    """

    @classmethod
    def name(cls) -> str:
        return "covid_deaths"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _file(self) -> str:
        return "covid_deaths_dataset.zip"

    @property
    def _record(self) -> str:
        return "4656009"

    @property
    def _prediction_length(self) -> int:
        return 30


@register_dataset
@dataclass(frozen=True)
class KddCupDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "KDD Cup 2018".
    """

    @classmethod
    def name(cls) -> str:
        return "kdd_2018"

    @property
    def max_training_time(self) -> int:
        return 14400

    @property
    def _file(self) -> str:
        return "kdd_cup_2018_dataset_without_missing_values.zip"

    @property
    def _record(self) -> str:
        return "4656756"

    @property
    def _prediction_length(self) -> int:
        return 48


@register_dataset
@dataclass(frozen=True)
class CifDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "CIF 2016".
    """

    @classmethod
    def name(cls) -> str:
        return "cif_2016"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def has_time_features(self) -> bool:
        return False

    @property
    def _file(self) -> str:
        return "cif_2016_dataset.zip"

    @property
    def _record(self) -> str:
        return "4656042"

    @property
    def _prediction_length(self) -> int:
        return 12


@register_dataset
@dataclass(frozen=True)
class AustralianElectricityDemandDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Australian Electricity Demand".
    """

    @classmethod
    def name(cls) -> str:
        return "australian_electricity_demand"

    @property
    def max_training_time(self) -> int:
        return 14400

    @property
    def _file(self) -> str:
        return "australian_electricity_demand_dataset.zip"

    @property
    def _record(self) -> str:
        return "4659727"

    @property
    def _prediction_length(self) -> int:
        return 48


@register_dataset
@dataclass(frozen=True)
class BitcoinDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Bitcoin".
    """

    @classmethod
    def name(cls) -> str:
        return "bitcoin"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _file(self) -> str:
        return "bitcoin_dataset_without_missing_values.zip"

    @property
    def _record(self) -> str:
        return "5122101"

    @property
    def _prediction_length(self) -> int:
        return 30


@register_dataset
@dataclass(frozen=True)
class RideshareDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Rideshare".
    """

    @classmethod
    def name(cls) -> str:
        return "rideshare"

    @property
    def max_training_time(self) -> int:
        return 14400

    @property
    def _file(self) -> str:
        return "rideshare_dataset_without_missing_values.zip"

    @property
    def _record(self) -> str:
        return "5122232"

    @property
    def _prediction_length(self) -> int:
        return 48


@register_dataset
@dataclass(frozen=True)
class VehicleTripsDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Vehicle Trips".
    """

    @classmethod
    def name(cls) -> str:
        return "vehicle_trips"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _file(self) -> str:
        return "vehicle_trips_dataset_without_missing_values.zip"

    @property
    def _record(self) -> str:
        return "5122537"

    @property
    def _prediction_length(self) -> int:
        return 30

    @property
    def _prediction_length_multiplier(self) -> int:
        return 1


@register_dataset
@dataclass(frozen=True)
class WeatherDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Weather".
    """

    @classmethod
    def name(cls) -> str:
        return "weather"

    @property
    def max_training_time(self) -> int:
        return 28800

    @property
    def _file(self) -> str:
        return "weather_dataset.zip"

    @property
    def _record(self) -> str:
        return "4654822"

    @property
    def _prediction_length(self) -> int:
        return 30


@register_dataset
@dataclass(frozen=True)
class TemperatureRainDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "Temperature Rain".
    """

    @classmethod
    def name(cls) -> str:
        return "temperature_rain"

    @property
    def max_training_time(self) -> int:
        return 28800

    @property
    def _file(self) -> str:
        return "temperature_rain_dataset_without_missing_values.zip"

    @property
    def _record(self) -> str:
        return "5129091"

    @property
    def _prediction_length(self) -> int:
        return 30


@register_dataset
@dataclass(frozen=True)
class M1YearlyDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "M1 Yearly".
    """

    @classmethod
    def name(cls) -> str:
        return "m1_yearly"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _file(self) -> str:
        return "m1_yearly_dataset.zip"

    @property
    def _record(self) -> str:
        return "4656193"

    @property
    def _prediction_length(self) -> int:
        return 6


@register_dataset
@dataclass(frozen=True)
class M1QuarterlyDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "M1 Quarterly".
    """

    @classmethod
    def name(cls) -> str:
        return "m1_quarterly"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _file(self) -> str:
        return "m1_quarterly_dataset.zip"

    @property
    def _record(self) -> str:
        return "4656154"

    @property
    def _prediction_length(self) -> int:
        return 8


@register_dataset
@dataclass(frozen=True)
class M1MonthlyDatasetConfig(MonashDatasetConfig):
    """
    The dataset configuration for "M1 Monthly".
    """

    @classmethod
    def name(cls) -> str:
        return "m1_monthly"

    @property
    def max_training_time(self) -> int:
        return 3600

    @property
    def _file(self) -> str:
        return "m1_monthly_dataset.zip"

    @property
    def _record(self) -> str:
        return "4656159"

    @property
    def _prediction_length(self) -> int:
        return 18


@register_dataset
@dataclass(frozen=True)
class RossmannDatasetConfig(KaggleDatasetConfig):
    """
    The dataset configuration for the "Rossmann Store Sales" Kaggle
    competition.
    """

    @classmethod
    def name(cls) -> str:
        return "rossmann"

    @property
    def max_training_time(self) -> int:
        return 7200

    @property
    def _link(self) -> str:
        return "https://www.kaggle.com/c/rossmann-store-sales"

    def _extract_data(
        self, path: Path
    ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
        # Read the raw data
        data = cast(pd.DataFrame, pd.read_csv(path / "train.csv"))
        stores = cast(pd.DataFrame, pd.read_csv(path / "store.csv"))

        # Generate GluonTS dataset
        metadata = {
            "freq": "D",
            "prediction_length": 48,
            "feat_static_cat": [
                {
                    "name": "store",
                    "cardinality": len(stores),
                },
            ],
        }

        series = []
        for i, store_data in data.groupby("Store"):
            sorted_data = store_data.sort_values("Date")
            series.append(
                {
                    "item_id": int(i) - 1,
                    "start": sorted_data.Date.min(),
                    "target": sorted_data.Sales.to_list(),
                    "feat_static_cat": [
                        int(i) - 1,
                    ],
                }
            )

        return metadata, series


@register_dataset
@dataclass(frozen=True)
class CorporacionFavoritaDatasetConfig(KaggleDatasetConfig):
    """
    The dataset configuration for the "Corporación Favorita" Kaggle
    competition.
    """

    @classmethod
    def name(cls) -> str:
        return "corporacion_favorita"

    @property
    def max_training_time(self) -> int:
        return 28800

    @property
    def _link(self) -> str:
        return "https://www.kaggle.com/c/favorita-grocery-sales-forecasting"

    def _extract_data(
        self, path: Path
    ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
        # Read the raw data
        data = cast(pd.DataFrame, pd.read_csv(path / "train.csv"))
        stores = cast(pd.DataFrame, pd.read_csv(path / "stores.csv"))
        item_ids = np.sort(data.item_nbr.unique())

        # Generate GluonTS dataset
        metadata = {
            "freq": "D",
            "prediction_length": 16,
            "feat_static_cat": [
                {
                    "name": "store",
                    "cardinality": len(stores),
                },
                {
                    "name": "item",
                    "cardinality": len(item_ids),
                },
            ],
        }

        series = []
        for i, ((item, store_id), group_data) in enumerate(
            data.groupby(["item_nbr", "store_nbr"])
        ):
            item_id = np.where(item_ids == item)[0][0]
            sorted_data = group_data.sort_values("date")
            sales = pd.Series(
                sorted_data.unit_sales.to_numpy(),
                index=pd.DatetimeIndex(sorted_data.date),
            )
            series.append(
                {
                    "item_id": i,
                    "start": sorted_data.date.min(),
                    "target": sales.resample("D")
                    .first()
                    .fillna(value=0)
                    .to_list(),
                    "feat_static_cat": [
                        int(store_id) - 1,
                        int(item_id),
                    ],
                }
            )

        return metadata, series


@register_dataset
@dataclass(frozen=True)
class WalmartDatasetConfig(KaggleDatasetConfig):
    """
    The dataset configuration for the "Walmart Recruiting" Kaggle competition.
    """

    @classmethod
    def name(cls) -> str:
        return "walmart"

    @property
    def max_training_time(self) -> int:
        return 7200

    @property
    def _link(self) -> str:
        return "https://www.kaggle.com/c/walmart-recruiting-store-sales-forecasting"

    def _extract_data(
        self, path: Path
    ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
        # Read the raw data
        data = cast(pd.DataFrame, pd.read_csv(path / "train.csv"))
        department_ids = np.sort(data.Dept.unique())

        # Generate GluonTS dataset
        metadata = {
            "freq": "W",
            "prediction_length": 39,
            "feat_static_cat": [
                {
                    "name": "store",
                    "cardinality": len(
                        data.Store.unique()
                    ),  # pylint: disable=no-member
                },
                {
                    "name": "department",
                    "cardinality": len(department_ids),
                },
            ],
        }

        series = []
        # pylint: disable=no-member
        for i, ((store_id, department), group_data) in enumerate(
            data.groupby(["Store", "Dept"])
        ):
            department_id = np.where(department_ids == department)[0][0]
            sorted_data = group_data.sort_values("Date")
            series.append(
                {
                    "item_id": i,
                    "start": sorted_data.Date.min(),
                    "target": sorted_data.Weekly_Sales.to_list(),
                    "feat_static_cat": [
                        int(store_id) - 1,
                        int(department_id),
                    ],
                }
            )

        return metadata, series


@register_dataset
@dataclass(frozen=True)
class RestaurantDatasetConfig(KaggleDatasetConfig):
    """
    The dataset configuration for the "Restaurant" Kaggle competition.
    """

    @classmethod
    def name(cls) -> str:
        return "restaurant"

    @property
    def max_training_time(self) -> int:
        return 7200

    @property
    def _link(self) -> str:
        return (
            "https://www.kaggle.com/c/recruit-restaurant-visitor-forecasting"
        )

    def _extract_data(
        self, path: Path
    ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
        # Read the raw data
        data = cast(pd.DataFrame, pd.read_csv(path / "air_visit_data.csv"))
        store_ids = np.sort(data.air_store_id.unique())

        # Generate GluonTS dataset
        metadata = {
            "freq": "D",
            "prediction_length": 39,
            "feat_static_cat": [
                {
                    "name": "restaurant",
                    "cardinality": len(store_ids),
                },
            ],
        }

        series = []
        # pylint: disable=no-member
        for i, (store, group_data) in enumerate(data.groupby("air_store_id")):
            store_id = np.where(store_ids == store)[0][0]
            sorted_data = group_data.sort_values("visit_date")
            visitors = pd.Series(
                sorted_data.visitors.to_numpy(),
                index=pd.DatetimeIndex(sorted_data.visit_date),
            )
            series.append(
                {
                    "item_id": i,
                    "start": sorted_data.visit_date.min(),
                    "target": visitors.resample("D")
                    .first()
                    .fillna(value=0)
                    .to_list(),
                    "feat_static_cat": [
                        int(store_id),
                    ],
                }
            )

        return metadata, series



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# pylint: disable=arguments-differ,too-many-ancestors,abstract-method
from typing import List, Tuple
import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import Callback, EarlyStopping
from torch import nn, optim
from .losses import ListMLELoss


class DeepSetLightningModule(pl.LightningModule):
    """
    Lightning module which trains a deep set model until convergence.
    """

    def __init__(
        self, model: nn.Module, loss: nn.Module, weight_decay: float = 0.0
    ):
        super().__init__()

        self.model = model
        self.loss = loss
        self.weight_decay = weight_decay
        self.uses_ranking = isinstance(self.loss, ListMLELoss)

    def configure_optimizers(self) -> optim.Optimizer:
        return optim.Adam(
            self.model.parameters(), lr=1e-2, weight_decay=self.weight_decay
        )

    def configure_callbacks(self) -> List[Callback]:
        return [
            EarlyStopping(
                "train_loss",
                patience=50,
                min_delta=1e-3,
                check_on_train_epoch_end=True,
            )
        ]

    def training_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
        _batch_idx: int,
    ) -> torch.Tensor:
        X, X_lengths, y_true, group_ids = batch
        y_pred = self.model(X, X_lengths)

        if self.uses_ranking:
            loss = self.loss(y_pred, y_true, group_ids)
        else:
            loss = self.loss(y_pred, y_true)

        self.log("train_loss", loss, on_step=False, on_epoch=True)
        return loss

    def predict_step(
        self, batch: Tuple[torch.Tensor, torch.Tensor], _batch_idx: int
    ) -> torch.Tensor:
        X, X_lengths = batch
        return self.model(X, X_lengths)
